# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from pyqtgraph import ViewBox
from qtpy.QtCore import QPoint, Qt, Signal
from qtpy.QtWidgets import QMenu

from .enums import MouseTool

ZOOM_IN = 1
ZOOM_OUT = -1


class KaraboViewBox(ViewBox):
    middleButtonClicked = Signal()

    def __init__(self, parent=None):
        super().__init__(parent, enableMenu=False)
        self.mouse_tool = MouseTool.Pointer
        self.setBackgroundColor("w")

        # Build a menu!
        self.menu = QMenu(parent)
        self.menu.addAction("View all", self.enableAutoRange)
        self.autorange_enabled = True

    # ---------------------------------------------------------------------
    # mouse events

    def mouseClickEvent(self, event):
        if event.button() == Qt.MiddleButton:
            if self.autorange_enabled:
                self.enableAutoRange()
            self.middleButtonClicked.emit()

        if self.mouse_tool is MouseTool.Zoom:
            if event.button() == Qt.LeftButton:
                self._click_zoom_mode(event, ZOOM_IN)
            elif event.button() == Qt.RightButton:
                self._click_zoom_mode(event, ZOOM_OUT)
        else:
            super().mouseClickEvent(event)

    def mouseDragEvent(self, event, axis=None):
        button = event.buttons()

        # Restore original MouseMode (and cursor) when dragEvent is finished
        if event.isFinish():
            super().mouseDragEvent(event, axis)
            self.set_mouse_tool(self.mouse_tool)
            return

        # Check on how the drag will commence
        if self.mouse_tool is MouseTool.Pointer and button == Qt.LeftButton:
            event.ignore()
            return
        elif self.mouse_tool is MouseTool.Move or button == Qt.MiddleButton:
            if event.isStart():
                # Enable panning regardless of MouseTool
                self.state["mouseMode"] = ViewBox.PanMode
                self.setCursor(Qt.ClosedHandCursor)
        elif button == Qt.RightButton:
            if event.isStart():
                # Change cursor on right button as well
                self.setCursor(Qt.ClosedHandCursor)

        super().mouseDragEvent(event, axis)

    def wheelEvent(self, event, axis=None):
        """Ignore mouse scroll since it also catches scene scroll"""
        event.ignore()

    # ---------------------------------------------------------------------
    # Public methods

    def set_mouse_tool(self, mode):
        if mode is MouseTool.Pointer:
            mouseMode = ViewBox.PanMode
            cursor = Qt.ArrowCursor
        elif mode is MouseTool.Zoom:
            mouseMode = ViewBox.RectMode
            cursor = Qt.CrossCursor
        elif mode is MouseTool.Move:
            mouseMode = ViewBox.PanMode
            cursor = Qt.OpenHandCursor
        else:
            raise ValueError(f"Invalid mouseTool: {mode}")

        # Using self.state to assign vb mouse mode to avoid emitting
        # sigStateChanged
        self.state["mouseMode"] = mouseMode
        self.setCursor(cursor)
        self.mouse_tool = mode

    def add_action(self, action, separator=True):
        if self.menu is None:
            return

        if separator:
            self.menu.addSeparator()
        self.menu.addAction(action)

    # ---------------------------------------------------------------------
    # Backward compatibility

    set_mouse_mode = set_mouse_tool

    @property
    def mouse_mode(self):
        return self.mouse_tool

    @mouse_mode.setter
    def mouse_mode(self, value):
        self.mouse_tool = value

    # ---------------------------------------------------------------------
    # Reimplemented `PyQtGraph` methods

    def removeItem(self, item):
        if item in self.allChildItems():
            super().removeItem(item)

    def menuEnabled(self):
        return self.menu is not None

    def raiseContextMenu(self, event):
        if self.menu is None:
            return
        pos = event.screenPos()
        self.menu.popup(QPoint(int(pos.x()), int(pos.y())))

    # ---------------------------------------------------------------------
    # private methods

    def _click_zoom_mode(self, event, direction):
        # Lifted from the PyQtGraph wheelEvent(). wheelScaleFactor can be
        # increased for larger scale/zoom

        scene_rect = self.sceneBoundingRect()
        pos = event.pos()
        if not scene_rect.contains(pos):
            event.ignore()
            return

        mask = np.array([1, 1])
        s = ((mask * 0.02) + 1) ** \
            (120 * direction * self.state["wheelScaleFactor"])

        center = self.mapToView(pos)
        self._resetTarget()
        self.scaleBy(s, center)
        self.sigRangeChangedManually.emit(self.state["mouseEnabled"])
        event.accept()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import re
from ast import literal_eval

import numpy as np
from qtpy.QtGui import QValidator
from traits.api import TraitError

from karabogui.binding.api import (
    VectorBoolBinding, VectorDoubleBinding, VectorFloatBinding,
    VectorInt8Binding, VectorInt16Binding, VectorInt32Binding,
    VectorInt64Binding, VectorStringBinding, VectorUint8Binding,
    VectorUint16Binding, VectorUint32Binding, VectorUint64Binding,
    convert_string, get_min_max, get_min_max_size)

BOOL_REGEX = r"(0|1|[T]rue|[F]alse)"
INT_REGEX = r"^[-+]?\d+$"
UINT_REGEX = r"^[+]?\d+$"
STRING_REGEX = r"^.+$"
# Was before r"^[-+]?\d*[\.]\d*$|^[-+]?\d+$"
DOUBLE_REGEX = r"^[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?$"

_CAST_MAP = {
    VectorInt8Binding: np.int8,
    VectorInt16Binding: np.int16,
    VectorInt32Binding: np.int32,
    VectorInt64Binding: np.int64,
    VectorUint8Binding: np.uint8,
    VectorUint16Binding: np.uint16,
    VectorUint32Binding: np.uint32,
    VectorUint64Binding: np.uint64,
}

_REGEX = {
    VectorBoolBinding: BOOL_REGEX,
    VectorDoubleBinding: DOUBLE_REGEX,
    VectorFloatBinding: DOUBLE_REGEX,
    VectorInt8Binding: INT_REGEX,
    VectorInt16Binding: INT_REGEX,
    VectorInt32Binding: INT_REGEX,
    VectorInt64Binding: INT_REGEX,
    VectorStringBinding: STRING_REGEX,
    VectorUint8Binding: UINT_REGEX,
    VectorUint16Binding: UINT_REGEX,
    VectorUint32Binding: UINT_REGEX,
    VectorUint64Binding: UINT_REGEX,
}

_INTERMEDIATE = {
    VectorBoolBinding: ("T", "t", "r", "u", "F", "a", "l", "s", ","),
    VectorDoubleBinding: ("+", "-", ",", "e"),
    VectorFloatBinding: ("+", "-", ",", ".", "e"),
    VectorInt8Binding: ("+", "-", ","),
    VectorInt16Binding: ("+", "-", ","),
    VectorInt32Binding: ("+", "-", ","),
    VectorInt64Binding: ("+", "-", ","),
    VectorStringBinding: (",",),
    VectorUint8Binding: ("+", ","),
    VectorUint16Binding: ("+", ","),
    VectorUint32Binding: ("+", ","),
    VectorUint64Binding: ("+", ","),
}


class ListValidator(QValidator):
    pattern = None
    intermediate = ()

    def __init__(self, binding, parent=None):
        super().__init__(parent=parent)
        self._binding = binding
        binding_type = type(self._binding)
        self.pattern = re.compile(_REGEX.get(binding_type, ""))
        self.intermediate = _INTERMEDIATE.get(binding_type, ())
        self.literal = (str if binding_type == VectorStringBinding
                        else literal_eval)
        # Note: Vector of integers are downcasted. We get a function
        # to validate that the user input matches the casted value
        self.cast = _CAST_MAP.get(binding_type, self.literal)

        minSize, maxSize = get_min_max_size(binding)
        self.minSize = minSize
        self.maxSize = maxSize

    def validate(self, input, pos):
        """The main validate function of the ListValidator"""
        # 1. Basic empty input check. Empty input might be acceptable!
        if input == "" and (self.minSize is None or not self.minSize):
            return self.Acceptable, input, pos
        elif input == "" and self.minSize is not None and self.minSize > 0:
            return self.Intermediate, input, pos
        elif input.startswith(","):
            return self.Intermediate, input, pos

        # 2. Check for size of list first
        values = [val for val in input.split(",")]
        if self.minSize is not None and len(values) < self.minSize:
            return self.Intermediate, input, pos
        if self.maxSize is not None and len(values) > self.maxSize:
            return self.Intermediate, input, pos

        # 3. Intermediate positions
        if input[-1] in self.intermediate:
            return self.Intermediate, input, pos

        # 4. Check every single list value if it is valid
        for value in values:
            if not self.pattern.match(value):
                return self.Invalid, input, pos
            # More checks including:
            # 4.1 We have to see we can cast it properly (SyntaxError)
            # 4.2 Check for changes in between Vectors (ValueError)
            # 4.3 Casts down integers (AssertionError), limits
            # 4.4 Overflowing integers (OverflowError)
            try:
                assert self.literal(value) == self.cast(value)
            except (ValueError, SyntaxError, AssertionError, OverflowError):
                return self.Intermediate, input, pos

        return self.Acceptable, input, pos


class BindingValidator(QValidator):
    """This is a soft generic binding validator

    Returns `Intermediate` if the binding cannot be validated by the trait. If
    the value can be validated by the binding trait it returns `Acceptable`.
    """

    def __init__(self, binding, parent=None):
        super().__init__(parent)
        self._binding = binding

    def validate(self, input, pos):
        """The main validate function using binding validate"""
        try:
            self._binding.validate_trait("value", input)
        except TraitError:
            return self.Intermediate, input, pos

        return self.Acceptable, input, pos


class SimpleValidator(QValidator):
    """This is a numeric binding validator which accounts min and max limits
    """

    def __init__(self, binding, parent=None):
        super().__init__(parent=parent)
        self._binding = binding
        self.minInc, self.maxInc = get_min_max(binding)

    def validate(self, input, pos):
        """Reimplemented function of QValidator to validate numeric input"""
        if input in ("+", "-", ""):
            return self.Intermediate, input, pos
        elif input[-1] in (" "):
            return self.Invalid, input, pos
        elif input[-1] in ("+", "-", "e"):
            return self.Intermediate, input, pos

        # Use the fast path validation
        try:
            self._binding.validate_trait("value", input)
        except TraitError:
            return self.Invalid, input, pos

        if self.inside_limits(input):
            return self.Acceptable, input, pos

        return self.Intermediate, input, pos

    def inside_limits(self, value):
        """Check if a value is within limits"""
        value, success = convert_string(value)
        if not success:
            return False
        if value < self.minInc or value > self.maxInc:
            return False

        # Check passed!
        return True



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os

from qtpy import uic
from qtpy.QtWidgets import QDialog


class TransformDialog(QDialog):
    def __init__(self, config, parent=None):
        super().__init__(parent)
        self.setModal(False)
        ui_path = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                               'transform_config.ui')
        uic.loadUi(ui_path, self)
        self.ui_offset.setValue(config['offset'])
        self.ui_step.setValue(config['step'])
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

    @property
    def transformations(self):
        config = {
            "offset": self.ui_offset.value(),
            "step": self.ui_step.value()}

        return config

    @staticmethod
    def get(configuration, parent=None):
        dialog = TransformDialog(configuration, parent)
        result = dialog.exec() == QDialog.Accepted
        content = {}
        content.update(dialog.transformations)

        return content, result



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest

from karabo.common.api import State
from karabo.common.scenemodel.api import DisplayStateColorModel
from karabo.native import Configurable, String
from karabogui.controllers.util import get_class_const_trait
from karabogui.indicators import STATE_COLORS
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..statecolor import DisplayStateColor


class Object(Configurable):
    state = String()


@pytest.fixture
def statecolor_setup(gui_app):
    # setup
    schema = Object.getClassSchema()
    model = DisplayStateColorModel()
    proxy = get_class_property_proxy(schema, "state")
    controller = DisplayStateColor(model=model, proxy=proxy)
    controller.create(None)
    assert controller.widget is not None

    yield controller, proxy

    # teardown
    controller.destroy()
    assert controller.widget is None


def test_statecolor_set_values(statecolor_setup):
    controller, proxy = statecolor_setup

    states = ("CHANGING", "ACTIVE", "PASSIVE", "DISABLED", "RUNNING",
              "STATIC", "NORMAL", "ERROR", "INIT", "UNKNOWN")
    for state in states:
        set_proxy_value(proxy, "state", state)
        color = STATE_COLORS[getattr(State, state)]
        sheet = controller.widget.styleSheet()
        assert str(color) in sheet


def test_statecolor_string(statecolor_setup):
    controller, proxy = statecolor_setup
    action = controller.widget.actions()[0]
    assert action.text() == "Show State String"
    state = "CHANGING"
    set_proxy_value(proxy, "state", state)
    assert controller.widget.text() == ""
    action.trigger()
    assert controller.widget.text() == "CHANGING"
    state = "ON"
    set_proxy_value(proxy, "state", state)
    assert controller.widget.text() == "ON"
    action.trigger()
    assert controller.widget.text() == ""


def test_statecolor_priority(statecolor_setup):
    controller, _ = statecolor_setup
    assert get_class_const_trait(controller, "_priority") == 90



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on March 11, 2019
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import traceback
from datetime import datetime

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QAction, QDialog, QFrame, QInputDialog, QLabel
from traits.api import Instance, Undefined

from karabo.common.scenemodel.api import DisplayTimeModel
from karabogui import messagebox
from karabogui.binding.api import BaseBinding, NodeBinding, get_binding_value
from karabogui.const import WIDGET_MIN_HEIGHT
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.dialogs.format_label import FormatLabelDialog
from karabogui.fonts import get_font_size_from_dpi
from karabogui.indicators import ALL_OK_COLOR
from karabogui.util import generateObjectName


def is_compatible(binding):
    return not isinstance(binding, NodeBinding)


@register_binding_controller(ui_name='Time Field',
                             klassname='TimeLabel',
                             binding_type=BaseBinding, priority=-10,
                             is_compatible=is_compatible,
                             can_show_nothing=False)
class DisplayTimeLabel(BaseBindingController):
    model = Instance(DisplayTimeModel, args=())

    def create_widget(self, parent):
        widget = QLabel(parent)
        widget.setAlignment(Qt.AlignCenter)
        widget.setMinimumHeight(WIDGET_MIN_HEIGHT)
        widget.setWordWrap(True)
        widget.setFrameStyle(QFrame.Box | QFrame.Plain)
        objectName = generateObjectName(self)
        widget.setObjectName(objectName)
        style_sheet = (f"QWidget#{objectName}" +
                       f" {{ background-color : rgba{ALL_OK_COLOR}; }}")
        widget.setStyleSheet(style_sheet)

        action = QAction('Change datetime format...', widget)
        widget.addAction(action)
        action.triggered.connect(self._change_time_format)

        action = QAction("Format field...", widget)
        action.triggered.connect(self._format_field)
        widget.addAction(action)

        self._apply_format(widget)
        return widget

    def value_update(self, proxy):
        if proxy.value is Undefined:
            return
        timestamp = proxy.binding.timestamp
        # Be backward compatible, some elements might not have a timestamp
        # in the past
        if timestamp is not None:
            dt = datetime.fromtimestamp(timestamp.toTimestamp())
            stamp = dt.strftime(self.model.time_format)
        else:
            stamp = "NaN"
        self.widget.setText(stamp)

    def _change_time_format(self):
        # NOTE: No extra protection required, as we do not allow altering
        # models for offline devices without binding
        text, ok = QInputDialog.getText(
            self.widget, 'Enter datetime format', 'datetime format = ',
            text=self.model.time_format)

        if not ok:
            return

        try:
            # Plainly try to apply the time_format on the value
            timestamp = self.proxy.binding.timestamp
            dt = datetime.fromtimestamp(timestamp.toTimestamp())
            dt.strftime(text)
        except Exception as e:
            err = traceback.format_exception_only(type(e), e)
            msg = '<pre>{1}{2}</pre>{3}'.format(*err)
            messagebox.show_error(msg, title='Error in datetime format',
                                  parent=self.widget)
            return
        self.model.time_format = text
        if get_binding_value(self.proxy) is not None:
            self.value_update(self.proxy)

    # -----------------------------------------------------------------------
    # Formatting methods

    def _format_field(self):
        dialog = FormatLabelDialog(font_size=self.model.font_size,
                                   font_weight=self.model.font_weight,
                                   parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            self.model.trait_set(font_size=dialog.font_size,
                                 font_weight=dialog.font_weight)
            self._apply_format()

    def _apply_format(self, widget=None):
        if widget is None:
            widget = self.widget

        # Apply font formatting
        font = widget.font()
        font.setPointSize(get_font_size_from_dpi(self.model.font_size))
        font.setBold(self.model.font_weight == "bold")
        widget.setFont(font)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest

from karabo.common.scenemodel.api import ScatterGraphModel
from karabo.native import Configurable, Double, Hash, Timestamp
from karabogui.binding.proxy import PropertyProxy
from karabogui.const import IS_WINDOWS_SYSTEM
from karabogui.testing import (
    get_class_property_proxy, set_proxy_hash, set_proxy_value)

from ..scatter_graph import DisplayScatterGraph


class Object(Configurable):
    x = Double()
    y = Double()


@pytest.fixture
def scatter_graph_setup(gui_app):
    schema = Object.getClassSchema()
    x = get_class_property_proxy(schema, "x")
    device = x.root_proxy
    y = PropertyProxy(root_proxy=device, path="y")
    controller = DisplayScatterGraph(proxy=x)
    controller.create(None)
    controller.visualize_additional_property(y)
    assert controller.widget is not None
    yield controller, x, y

    controller.destroy()
    assert controller.widget is None


@pytest.mark.skipif(IS_WINDOWS_SYSTEM,
                    reason="curve.getData returns empty arrays in Windows")
def test_scatter_graph_basics(scatter_graph_setup):
    controller, x, y = scatter_graph_setup
    set_proxy_value(x, "x", 2.1)
    set_proxy_value(y, "y", 3.2)
    curve = controller._plot
    assert list(curve.getData()) == [2.1, 3.2]


@pytest.mark.skipif(IS_WINDOWS_SYSTEM,
                    reason="curve.getData returns empty arrays in Windows")
def test_scatter_timestamp(scatter_graph_setup):
    controller, x, y = scatter_graph_setup
    timestamp = Timestamp()
    set_proxy_value(x, "x", 2.1)
    h = Hash("y", 3.2)
    set_proxy_hash(y, h, timestamp)
    curve = controller._plot
    assert list(curve.getData()) == [2.1, 3.2]
    h = Hash("y", 13.2)
    set_proxy_hash(y, h, timestamp)
    assert list(curve.getData()) == [2.1, 3.2]


def test_deque(mocker, scatter_graph_setup):
    controller, x, _ = scatter_graph_setup
    controller = DisplayScatterGraph(proxy=x, model=ScatterGraphModel())
    controller.create(None)
    action = controller.widget.actions()[10]
    assert action.text() == "Queue Size"

    path = "karabogui.controllers.display.scatter_graph.QInputDialog.getInt"
    mocker.patch(path, return_value=(20, True))
    action.trigger()
    assert controller.model.maxlen == 20
    assert controller._x_values.maxlen == 20
    assert controller._y_values.maxlen == 20


def test_pointsize(mocker, scatter_graph_setup):
    controller, x, _ = scatter_graph_setup
    controller = DisplayScatterGraph(proxy=x, model=ScatterGraphModel())
    controller.create(None)
    action = controller.widget.actions()[11]
    assert action.text() == "Point Size"

    path = "karabogui.controllers.display.scatter_graph.QInputDialog.getDouble"
    mocker.patch(path, return_value=(2.7, True))
    action.trigger()
    assert controller.model.psize == 2.7
    assert controller._plot.opts["size"] == 2.7


def test_removal_property(scatter_graph_setup):
    controller, x, y = scatter_graph_setup
    controller = DisplayScatterGraph(proxy=x, model=ScatterGraphModel())
    controller.create(None)
    controller.visualize_additional_property(y)
    # Put a few values
    set_proxy_value(x, "x", 2.1)
    set_proxy_value(y, "y", 3.2)
    set_proxy_value(x, "x", 5.3)
    set_proxy_value(y, "y", 4.6)

    if not IS_WINDOWS_SYSTEM:
        assert controller._last_x_value is not None
        assert len(controller._x_values) > 0
        assert len(controller._y_values) > 0

    assert controller.remove_additional_property(y)
    assert controller._y_proxy is None
    assert controller._last_x_value is None
    assert len(controller._x_values) == 0
    assert len(controller._y_values) == 0

    assert not controller.remove_additional_property(y)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from qtpy.QtCore import QPointF

from karabogui.graph.common.api import (
    CrosshairROI, ImageROIController, RectROI, ROITool)
from karabogui.graph.image.plot import KaraboImagePlot
from karabogui.testing import GuiTestCase


class TestKaraboImagePlotROIController(GuiTestCase):

    def setUp(self):
        super().setUp()

        self.plot_item = KaraboImagePlot()
        self.roi_controller = ImageROIController(self.plot_item)
        self.roi_controller.enable_updates(True)
        self.roi_controller.updated.connect(self._mock_emit)

        self._region = None

        assert self.roi_controller.current_tool_item is None
        assert self.roi_controller.current_tool == ROITool.NoROI

    def _mock_emit(self, value):
        self._region = value

    def test_image_rect_roi(self):
        roi = self.roi_controller.add(ROITool.Rect,
                                      pos=QPointF(-3, -3),
                                      size=(3, 3))

        image = np.zeros((125, 125))

        assert roi.center == QPointF(-1.5, -1.5)
        assert roi.textItem.toPlainText() == (
            "Region of Interest\nCenter: (-1.5, -1.5)\nSize: (3, 3)")

        # 1. Check invalid ROI (out of bounds)
        self.plot_item.set_image(image)
        x_axis, y_axis = self.plot_item.transformed_axes
        assert not self._region.valid()

        # 2. Check full intersection ROI
        roi.setPos(QPointF(15, 15))
        assert self._region.valid()
        x_slice, y_slice = self._region.slices
        assert np.array_equal(x_axis[x_slice], [15, 16, 17])
        assert np.array_equal(y_axis[y_slice], [15, 16, 17])

        # 3. Check partial intersection
        roi.setPos(QPointF(-1, -1))
        assert self._region.valid()
        x_slice, y_slice = self._region.slices
        assert np.array_equal(x_axis[x_slice], [0, 1])
        assert np.array_equal(y_axis[y_slice], [0, 1])

        assert isinstance(self.roi_controller.current_tool_item, RectROI)
        assert self.roi_controller.current_tool == ROITool.Rect

    def test_image_crosshair_roi(self):
        roi = self.roi_controller.add(ROITool.Crosshair,
                                      pos=QPointF(-3, -3))

        image = np.zeros((125, 125))

        # ROI is outside, so should emit invalid ImageRegion
        self.plot_item.set_image(image)
        assert not self._region.valid()

        # Move the ROI to a valid x position (projects y-axis)
        roi.setPos(QPointF(-3, 3))
        assert roi.center == QPointF(-3.000000, 3.000000)
        assert roi.textItem.toPlainText() == \
               "Region of Interest\nCenter: (-3, 3)"

        self.plot_item.set_image(image)
        x_axis = self.plot_item.transformed_axes[0]
        x_slice = self._region.slices[0]

        assert not self._region.valid(axis=1)
        assert self._region.valid(axis=0)
        assert np.array_equal(x_axis[x_slice], np.arange(125))

        # Move the ROI to a valid y position (projects x-axis)
        roi.setPos(QPointF(3, -3))
        assert roi.center == QPointF(3.000000, -3.000000)
        assert roi.textItem.toPlainText() == (
            "Region of Interest\nCenter: (3, -3)")
        assert roi.coords == (3.000000, -3.000000)
        self.plot_item.set_image(image)
        y_axis = self.plot_item.transformed_axes[1]
        y_slice = self._region.slices[1]

        assert not self._region.valid(axis=0)
        assert self._region.valid(axis=1)
        assert np.array_equal(y_axis[y_slice], np.arange(125))

        assert isinstance(self.roi_controller.current_tool_item, CrosshairROI)
        assert self.roi_controller.current_tool == ROITool.Crosshair



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtCore import QPoint
from qtpy.QtWidgets import QMessageBox

from karabogui.dialogs.api import KaraboMessageBox
from karabogui.testing import GuiTestCase


class TestMessageBox(GuiTestCase):

    def test_basic(self):
        """Check the basics of the Karabo Messagebox"""
        box_error = self._create_messagebox(
            QMessageBox.Critical,
            "Caution: Pump broke",
            "Temperature to high")

        box_info = self._create_messagebox(
            QMessageBox.Information,
            "New device online",
            "The new motor arrived in the experimental hall")

        assert len(box_info.existing) == 2
        assert len(box_error.existing) == 2

        m = "Temperature to high"
        assert box_error.details_textedit.toPlainText() == m
        assert box_error._is_showing_details is False
        self.click(box_error.show_details_button)
        self.process_qt_events()
        assert box_error._is_showing_details is True
        assert box_info.text_label.text() == "New device online"
        box_error.close()
        assert len(box_error.existing) == 1
        assert len(box_info.existing) == 1

        # Close same again, graceful
        box_error.close()

        box_warning = self._create_messagebox(
            QMessageBox.Warning,
            "Take care!")
        assert len(box_warning.existing) == 2

        with mock.patch("karabogui.dialogs.messagebox.QMenu") as m:
            box_warning._show_context_menu(QPoint(0, 0))
            m.assert_called_once()

    def _create_messagebox(self, icon, text, detail=None):
        """Create a messagebox with `icon`, `text` and `detail`"""
        mbox = KaraboMessageBox()
        mbox.setIcon(icon)
        mbox.setText(text)
        if detail is not None:
            mbox.setDetailedText(detail)
        mbox.update()
        return mbox


if __name__ == "__main__":
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np

import karabogui.binding.binding_types as types
from karabo.common.const import (
    KARABO_SCHEMA_DEFAULT_VALUE, KARABO_SCHEMA_MAX_EXC, KARABO_SCHEMA_MAX_INC,
    KARABO_SCHEMA_MAX_SIZE, KARABO_SCHEMA_MIN_EXC, KARABO_SCHEMA_MIN_INC,
    KARABO_SCHEMA_MIN_SIZE, KARABO_SCHEMA_OPTIONS)
from karabo.native import (
    AccessMode, Bool, Configurable, Hash, HashList, String, UInt8)
from karabogui.binding.builder import build_binding
from karabogui.binding.validate import (
    convert_string, get_default_value, sanitize_table_value,
    validate_binding_configuration, validate_table_value, validate_value)
from karabogui.testing import get_simple_props_schema


class TableRow(Configurable):
    stringProperty = String(defaultValue='foo')
    uintProperty = UInt8(defaultValue=1)
    boolProperty = Bool(defaultValue=True)


class TableRowEmpty(Configurable):
    stringProperty = String()
    uintProperty = UInt8()
    boolProperty = Bool()


class TableRowReadOnly(Configurable):
    stringProperty = String(accessMode=AccessMode.READONLY)


class TableRowReadOnlyDefault(Configurable):
    uintProperty = UInt8(defaultValue=57,
                         accessMode=AccessMode.READONLY)


class TableRowMixedDefaults(Configurable):
    stringProperty = String(defaultValue='bar')
    uintProperty = UInt8(defaultValue=20)
    boolProperty = Bool()


def test_validate_value_float():
    """Test float binding validation"""
    # Check valid values
    binding = types.FloatBinding()
    assert_binding(binding, 1.3, expected=1.3)
    assert_binding(binding, np.float32(1.3), expected=np.float32(1.3))
    assert_binding(binding, np.float64(1.3), expected=np.float64(1.3))
    assert_binding(binding, np.uint32(1), expected=float(1.0))
    assert_binding(binding, np.int32(-1), expected=float(-1.0))
    assert_binding(binding, "1.0", expected=float(1.0))

    # Check invalid values for floats
    assert_binding(binding, "foo", valid=False)
    assert_binding(binding, [1, 2, 3], valid=False)


def assert_binding(binding, value, expected=None, valid=True):
    """Validate a value against its binding

    :param binding: The corresponding binding
    :param value: The value to be verified
    :param expected: Expected value after validation. Must be provided
                     on `valid` case.
    :param valid: If we have a `valid` value assert. Default is `True`.
    """
    validated = validate_value(binding, value)
    if not valid:
        assert validated is None
    else:
        # Always force to specify expected
        assert expected is not None
        assert validated == expected
        assert issubclass(type(validated), type(expected))


def test_validate_value_uint8():
    """Test uint8 binding validation."""
    binding = types.Uint8Binding()

    # Check valid values
    assert_binding(binding, 0, expected=np.uint8(0))
    assert_binding(binding, 255, expected=np.uint8(255))
    assert_binding(binding, np.int8(8), expected=np.uint8(8))
    assert_binding(binding, '1', expected=np.uint8(1))

    # Check invalid values
    assert_binding(binding, -1, valid=False)
    assert_binding(binding, 256, valid=False)
    assert_binding(binding, 1.0, valid=False)  # the value is a float
    assert_binding(binding, "foo", valid=False)
    assert_binding(binding, [1, 2, 3], valid=False)
    assert_binding(binding, HashList(), valid=False)


def test_validate_value_int8():
    """Test int8 binding validation."""
    binding = types.Int8Binding()

    # Check valid values
    assert_binding(binding, -128, expected=np.int8(-128))
    assert_binding(binding, 127, expected=np.int8(127))
    assert_binding(binding, '1', expected=np.int8(1))
    assert_binding(binding, np.uint32(1), expected=np.int8(1))

    # Check invalid values
    assert_binding(binding, -129, valid=False)
    assert_binding(binding, 128, valid=False)
    assert_binding(binding, 1.0, valid=False)  # the value is a float
    assert_binding(binding, np.float64(1.0), valid=False)
    assert_binding(binding, "foo", valid=False)
    assert_binding(binding, [1, 2, 3], valid=False)
    assert_binding(binding, HashList([Hash(), Hash()]), valid=False)


def test_validate_value_int16():
    """Test the int16 binding validation."""
    binding = types.Int16Binding()

    # Check valid values
    assert_binding(binding, 0, expected=np.int16(0))
    assert_binding(binding, (2 ** 15) - 1, expected=np.int16((2 ** 15) - 1))
    assert_binding(binding, np.int8(8), expected=np.int16(8))
    assert_binding(binding, '1', expected=np.int16(1))

    # Check invalid values
    assert_binding(binding, -(2 ** 64), valid=False)
    assert_binding(binding, 1.0, valid=False)  # the value is a float
    assert_binding(binding, "foo", valid=False)
    assert_binding(binding, [1, 2, 3], valid=False)
    assert_binding(binding, HashList(), valid=False)


def test_validate_value_uint16():
    """Test uint16 binding validation."""
    binding = types.Uint16Binding()

    # Check valid values
    assert_binding(binding, 0, expected=np.uint16(0))
    assert_binding(binding, (2 ** 16) - 1, expected=np.uint16((2 ** 16) - 1))
    assert_binding(binding, np.int8(8), expected=np.uint16(8))
    assert_binding(binding, '1', expected=np.uint16(1))

    # Check invalid values
    assert_binding(binding, -1, valid=False)
    assert_binding(binding, (2 ** 16), valid=False)
    assert_binding(binding, 1.0, valid=False)  # the value is a float
    assert_binding(binding, "foo", valid=False)
    assert_binding(binding, [1, 2, 3], valid=False)
    assert_binding(binding, HashList(), valid=False)


def test_validate_value_int32():
    """Test int32 binding validation."""
    binding = types.Int32Binding()

    # Check valid values
    assert_binding(binding, -(2 ** 31), expected=np.int32(-(2 ** 31)))
    assert_binding(binding, (2 ** 31) - 1, expected=np.int32((2 ** 31) - 1))
    assert_binding(binding, '1', expected=np.int32(1))
    assert_binding(binding, np.uint32(1), expected=np.int32(1))

    # Check invalid values
    assert_binding(binding, -(2 ** 32), valid=False)
    assert_binding(binding, (2 ** 32) - 1, valid=False)
    assert_binding(binding, 1.0, valid=False)  # the value is a float
    assert_binding(binding, np.float64(1.0), valid=False)
    assert_binding(binding, "foo", valid=False)
    assert_binding(binding, [1, 2, 3], valid=False)
    assert_binding(binding, HashList([Hash(), Hash()]), valid=False)


def test_validate_value_uint32():
    """Test uint32 binding validation."""
    binding = types.Uint32Binding()

    # Check valid values
    assert_binding(binding, 1, expected=np.uint32(1))
    assert_binding(binding, (2 ** 32) - 1, expected=np.uint32((2 ** 32) - 1))
    assert_binding(binding, '1', expected=np.uint32(1))
    assert_binding(binding, np.uint32(1), expected=np.uint32(1))

    # Check invalid values
    assert_binding(binding, -(2 ** 32), valid=False)
    assert_binding(binding, (2 ** 32), valid=False)
    assert_binding(binding, 1.0, valid=False)  # the value is a float
    assert_binding(binding, np.float64(1.0), valid=False)
    assert_binding(binding, "foo", valid=False)
    assert_binding(binding, [1, 2, 3], valid=False)
    assert_binding(binding, HashList([Hash(), Hash()]), valid=False)


def test_validate_value_int64():
    """Test int64 binding validation."""
    binding = types.Int64Binding()

    # Check valid values
    assert_binding(binding, -(2 ** 63), expected=np.int64(-(2 ** 63)))
    assert_binding(binding, (2 ** 63) - 1, expected=np.int64((2 ** 63) - 1))
    assert_binding(binding, '1', expected=np.int64(1))
    assert_binding(binding, np.uint32(1), expected=np.int64(1))

    # Check invalid values
    assert_binding(binding, -(2 ** 64), valid=False)
    assert_binding(binding, (2 ** 64) - 1, valid=False)
    assert_binding(binding, 1.0, valid=False)  # the value is a float
    assert_binding(binding, np.float64(1.0), valid=False)
    assert_binding(binding, "foo", valid=False)
    assert_binding(binding, [1, 2, 3], valid=False)
    assert_binding(binding, HashList([Hash(), Hash()]), valid=False)


def test_validate_value_uint64():
    """Test uint64 binding validation."""
    binding = types.Uint64Binding()

    # Check valid values
    assert_binding(binding, 1, expected=np.uint64(1))
    assert_binding(binding, (2 ** 64) - 1, expected=np.uint64((2 ** 64) - 1))
    assert_binding(binding, '1', expected=np.uint64(1))
    assert_binding(binding, np.uint32(1), expected=np.uint64(1))

    # Check invalid values
    assert_binding(binding, -(2 ** 128), valid=False)
    assert_binding(binding, (2 ** 128) - 1, valid=False)
    assert_binding(binding, 1.0, valid=False)  # the value is a float
    assert_binding(binding, np.float64(1.0), valid=False)
    assert_binding(binding, "foo", valid=False)
    assert_binding(binding, [1, 2, 3], valid=False)
    assert_binding(binding, HashList([Hash(), Hash()]), valid=False)


def test_validate_min_max_binding():
    """Test the validate with min and max attributes of a binding"""
    attrs = {KARABO_SCHEMA_MIN_EXC: 0, KARABO_SCHEMA_MAX_EXC: 4}
    binding = types.Int8Binding(attributes=attrs)
    assert_binding(binding, 0, valid=False)
    assert_binding(binding, -1, valid=False)
    assert_binding(binding, 3, expected=np.int8(3))
    assert_binding(binding, '1', expected=np.int8(1))
    assert_binding(binding, 4, valid=False)

    attrs = {KARABO_SCHEMA_MIN_INC: 0, KARABO_SCHEMA_MAX_INC: 4}
    binding = types.Int8Binding(attributes=attrs)
    assert_binding(binding, 0, expected=np.int8(0))
    assert_binding(binding, 4, expected=np.int8(4))
    assert_binding(binding, '1', expected=np.int8(1))

    # And the float
    attrs = {KARABO_SCHEMA_MIN_EXC: 0, KARABO_SCHEMA_MAX_EXC: 4}
    binding = types.FloatBinding(attributes=attrs)
    assert_binding(binding, 0.0, valid=False)
    assert_binding(binding, 0.00000000000001, expected=float(0.00000000000001))
    assert_binding(binding, -4.7, valid=False)
    assert_binding(binding, 3.99999999999999, expected=float(3.99999999999999))
    assert_binding(binding, 4.00000000000000000000000000000001, valid=False)

    attrs = {KARABO_SCHEMA_MIN_INC: 0, KARABO_SCHEMA_MAX_INC: 4}
    binding = types.FloatBinding(attributes=attrs)
    assert_binding(binding, 0, expected=float(0.0))
    assert_binding(binding, 4, expected=float(4.0))
    assert_binding(binding, '1', expected=float(1.0))


def test_validate_value_string():
    """Test string binding validation.
    If valid, it should return the casted value"""
    binding = types.StringBinding()

    # Check valid values
    assert validate_value(binding, "1") == "1"
    assert validate_value(binding, 1) == "1"
    assert validate_value(binding, 1.0) == "1.0"

    # Check invalid values
    assert validate_value(binding, [1, 2, 3]) is None
    assert validate_value(binding, HashList()) is None


def test_validate_value_bool():
    """Test bool binding validation.
    If valid, it should return the casted value. The casting considers the
    the __bool__ of the object (the conventional way), not sure if it's what
    we want in Karabo."""
    binding = types.BoolBinding()

    # Check valid values
    assert validate_value(binding, 1) is True
    assert validate_value(binding, 0) is False
    assert validate_value(binding, 1.0) is True
    assert validate_value(binding, 0) is False
    assert validate_value(binding, '1') is True
    assert validate_value(binding, '0') is False
    assert validate_value(binding, "True") is True
    assert validate_value(binding, "False") is False

    # Check dubious values
    assert validate_value(binding, [1, 2, 3]) is True  # this should be invalid
    assert validate_value(binding, 2.0) is True  # this should be invalid
    assert validate_value(binding, HashList([Hash(), Hash()])) is True


def test_validate_value_vectoruint8():
    binding = types.VectorUint8Binding()

    array = np.array([0, 10, 255])
    # Check valid values
    _assert_array(binding, array, dtype=np.uint8)
    _assert_array(binding, array.astype(np.float32), dtype=np.uint8)
    _assert_array(binding, array.astype(np.uint32), dtype=np.uint8)

    # Check invalid values
    assert validate_value(binding, [-1]) is None
    assert validate_value(binding, [256]) is None
    assert validate_value(binding, 0) is None
    assert validate_value(binding, 255) is None
    assert validate_value(binding, 1.0) is None
    assert validate_value(binding, "foo") is None
    assert validate_value(binding, HashList([Hash(), Hash()])) is None


def test_validate_value_vectorint8():
    binding = types.VectorInt8Binding()

    # Check valid values
    _assert_array(binding, [-1, 0, 1], dtype=np.int8)
    _assert_array(binding, [-128, 0, 127], dtype=np.int8)
    _assert_array(binding, [-1.0, 0.0, 1.0], dtype=np.int8)

    # Check invalid values
    assert validate_value(binding, [129]) is None
    assert validate_value(binding, [128]) is None
    assert validate_value(binding, -128) is None
    assert validate_value(binding, 127) is None
    assert validate_value(binding, 1.0) is None
    assert validate_value(binding, "foo") is None
    assert validate_value(binding, HashList([Hash(), Hash()])) is None


def test_validate_value_vectorfloat():
    binding = types.VectorFloatBinding()
    array = np.array([-1, 0, 1])

    # Check valid values
    _assert_array(binding, array, dtype=np.float32)
    _assert_array(binding, array.astype(np.float32), dtype=np.float32)
    _assert_array(binding, array.astype(np.float64), dtype=np.float32)

    # Check invalid values
    assert validate_value(binding, -129) is None
    assert validate_value(binding, 128) is None
    assert validate_value(binding, 1.0) is None
    assert validate_value(binding, "foo") is None
    assert validate_value(binding, HashList([Hash(), Hash()])) is None


def test_validate_value_vectordouble():
    binding = types.VectorDoubleBinding()
    array = np.array([-1, 0, 1])

    # Check valid values
    _assert_array(binding, array, dtype=np.float64)
    _assert_array(binding, array.astype(np.float32), dtype=np.float64)
    _assert_array(binding, array.astype(np.float64), dtype=np.float64)

    # Check invalid values
    assert validate_value(binding, -129) is None
    assert validate_value(binding, 128) is None
    assert validate_value(binding, 1.0) is None
    assert validate_value(binding, "foo") is None
    assert validate_value(binding, HashList([Hash(), Hash()])) is None


def test_validate_value_vector_size():
    attributes = {KARABO_SCHEMA_MIN_SIZE: 3, KARABO_SCHEMA_MAX_SIZE: 5}
    binding = types.VectorDoubleBinding(attributes=attributes)
    array = np.array([-1, 0, 1])
    _assert_array(binding, array, dtype=np.float64)
    array = np.array([-1, 0, 1, 25, 100])
    _assert_array(binding, array, dtype=np.float64)

    # Check valid values, minimum size
    array = np.array([-1, 0])
    assert validate_value(binding, array) is None

    # Check valid values, maximum size
    array = np.array([-1, 0, 1, 25, 100, 129])
    assert validate_value(binding, array) is None


def test_validate_value_options():
    attributes = {KARABO_SCHEMA_OPTIONS: [3, 4, 5]}
    binding = types.Uint8Binding(attributes=attributes)
    assert validate_value(binding, "3") == 3
    assert validate_value(binding, 4) == 4
    # Not in options!
    assert validate_value(binding, 2) is None
    assert validate_value(binding, 6) is None

    attributes = {KARABO_SCHEMA_OPTIONS: ["ham", "eggs"]}
    binding = types.StringBinding(attributes=attributes)
    assert validate_value(binding, "ham") == "ham"
    assert validate_value(binding, "eggs") == "eggs"
    assert validate_value(binding, 2) is None
    assert validate_value(binding, "XFEL") is None


def test_validate_vector_hash():
    """Test vector hash binding validation.
    It returns both valid and invalid values in the form of a HashList([]) for
    each row."""

    binding = types.VectorHashBinding(
        row_schema=TableRow.getClassSchema().hash)
    empty_binding = types.VectorHashBinding(
        row_schema=TableRowEmpty.getClassSchema().hash)

    # Check with one valid hash
    valid_hash = Hash("stringProperty", "foo",
                      "uintProperty", 1,
                      "boolProperty", True)
    valid_hashlist = HashList([valid_hash])
    valid, invalid = validate_table_value(binding, valid_hashlist)
    assert valid == valid_hashlist
    assert invalid == HashList([])

    # Check two valid hashes
    valid_hash = Hash("stringProperty", "foo",
                      "uintProperty", 1,
                      "boolProperty", True)
    valid_hashlist = HashList([valid_hash, valid_hash])
    valid, invalid = validate_table_value(binding, valid_hashlist)
    assert valid == valid_hashlist
    assert invalid == HashList([])

    # Check with one hash that contains an invalid property
    invalid_hash = Hash("stringProperty", "foo",
                        "uintProperty", -1,  # invalid for a uintProperty
                        "boolProperty", True)
    valid, invalid = validate_table_value(binding, HashList([invalid_hash]))
    assert valid == HashList([])
    assert invalid == HashList([invalid_hash])

    # Check with two hashes which one hash contains invalid properties
    valid_hash = Hash("stringProperty", "foo",
                      "uintProperty", 1,
                      "boolProperty", True)
    invalid_hash = Hash("stringProperty", "foo",
                        "uintProperty", -1,  # invalid for a uintProperty
                        "boolProperty", True)
    hash_list = HashList([valid_hash, invalid_hash])
    valid, invalid = validate_table_value(binding, hash_list)
    assert valid == HashList([valid_hash])  # drop the row with invalids
    assert invalid == HashList([invalid_hash])

    # Check with two hashes which both contains invalid properties
    valid_hash = Hash("stringProperty", [1, 2],  # invalid for a uintProperty
                      "uintProperty", 1,
                      "boolProperty", "bar")
    invalid_hash = Hash("stringProperty", "foo",
                        "uintProperty", -1,  # invalid for a uintProperty
                        "boolProperty", True)
    hash_list = HashList([valid_hash, invalid_hash])
    valid, invalid = validate_table_value(binding, hash_list)
    assert valid == HashList([])  # drop the row with invalids
    assert invalid == HashList([valid_hash, invalid_hash])

    # Check with a hash that contains an invalid property and default value
    invalid_hash = Hash("stringProperty", "foo",
                        "boolProperty", True)
    valid, invalid = validate_table_value(binding, HashList([invalid_hash]))
    assert valid == HashList([Hash("stringProperty", "foo",
                                   "uintProperty", 1,  # default value
                                   "boolProperty", True)])
    assert invalid == HashList([])

    # Check with a hash that contains an invalid property
    # and without default value
    invalid_hash = Hash("stringProperty", "foo",
                        "boolProperty", True)
    valid, invalid = validate_table_value(empty_binding,
                                          HashList([invalid_hash]))
    assert valid == HashList([])
    assert invalid == HashList([Hash("stringProperty", "foo",
                                     "uintProperty", None,  # dropped val
                                     "boolProperty", True)])

    # ReadOnly tables should be still validated
    # -----------------------------------------------------------------------

    # 1. A fully valid Hash
    readonly_binding = types.VectorHashBinding(
        row_schema=TableRowReadOnly.getClassSchema().hash)
    readonly_binding_default = types.VectorHashBinding(
        row_schema=TableRowReadOnlyDefault.getClassSchema().hash)

    valid_hash = Hash("stringProperty", "karabo")
    valid, invalid = validate_table_value(readonly_binding,
                                          HashList([valid_hash]))
    assert valid == HashList([Hash("stringProperty", "karabo")])
    assert invalid == HashList([])

    # 2. A valid string but boolean is not in the schema

    invalid_hash = Hash("stringProperty", "foo",
                        "boolProperty", True)  # A value to many
    valid, invalid = validate_table_value(readonly_binding,
                                          HashList([invalid_hash]))
    assert valid == HashList([Hash("stringProperty", "foo")])
    assert invalid == HashList([])

    # 3. An empty Hash without default value in Schema
    invalid_hash = Hash()
    valid, invalid = validate_table_value(readonly_binding,
                                          HashList([invalid_hash]))
    assert valid == HashList([])
    assert invalid == HashList([Hash("stringProperty", None)])

    # 3. An empty Hash WITH default value in Schema
    invalid_hash = Hash()
    valid, invalid = validate_table_value(readonly_binding_default,
                                          HashList([invalid_hash]))
    assert valid == HashList([Hash("uintProperty", 57)])
    assert invalid == HashList([])


def test_sanitize_table():
    """Test vector hash binding sanitization."""
    binding = types.VectorHashBinding(
        row_schema=TableRow.getClassSchema().hash)
    empty_binding = types.VectorHashBinding(
        row_schema=TableRowEmpty.getClassSchema().hash)
    readonly_binding = types.VectorHashBinding(
        row_schema=TableRowReadOnly.getClassSchema().hash)
    mixed_default_binding = types.VectorHashBinding(
        row_schema=TableRowMixedDefaults.getClassSchema().hash)

    VALID_TABLE_HASH = Hash(
        "stringProperty", "foo",
        "uintProperty", 1,
        "boolProperty", True)

    # Check with a single hash conform to the schema
    hash_list = HashList([VALID_TABLE_HASH])
    success, value = sanitize_table_value(binding, hash_list)
    assert value == hash_list
    assert success is True

    hash_list = HashList([VALID_TABLE_HASH, VALID_TABLE_HASH])
    success, value = sanitize_table_value(binding, hash_list)
    assert value == hash_list
    assert success is True

    # Check one hash that contains an invalid value for a property
    invalid_hash = Hash(
        "stringProperty", "foo",
        "uintProperty", -1,  # invalid for a uintProperty
        "boolProperty", True)

    success, value = sanitize_table_value(binding, HashList([invalid_hash]))
    assert value == HashList([VALID_TABLE_HASH])
    assert success is False

    # Check with two hashes which one hash contains invalid properties
    invalid_hash = Hash(
        "stringProperty", "foo",
        "uintProperty", -1,  # invalid for a uintProperty
        "boolProperty", True)
    hash_list = HashList([VALID_TABLE_HASH, invalid_hash])
    success, value = sanitize_table_value(binding, hash_list)
    assert value == HashList([VALID_TABLE_HASH, VALID_TABLE_HASH])
    assert success is False

    # Check with two hashes which both contains invalid properties
    invalid_hash_1 = Hash(
        "stringProperty", [1, 2],  # invalid for a uintProperty
        "uintProperty", 1,
        "boolProperty", "bar")
    invalid_hash_2 = Hash(
        "stringProperty", "foo",
        "uintProperty", -1,  # invalid for a uintProperty
        "boolProperty", True)
    hash_list = HashList([invalid_hash_1, invalid_hash_2])
    success, value = sanitize_table_value(binding, hash_list)
    assert value == HashList([VALID_TABLE_HASH, VALID_TABLE_HASH])
    assert success is False

    # Check with a hash that misses a property. The schema has a default value
    invalid_hash = Hash("stringProperty", "foo",
                        "boolProperty", True)
    success, value = sanitize_table_value(binding, HashList([invalid_hash]))
    assert value == HashList([VALID_TABLE_HASH])
    assert success is False

    # Check with a hash that has false property keys.
    # The schema has a default values in this test case
    invalid_hash = Hash("stringProperty", "foo",
                        "uintPropertyWRONG", 1,
                        "boolPropertyWRONG", True)
    success, value = sanitize_table_value(binding, HashList([invalid_hash]))
    assert value == HashList([VALID_TABLE_HASH])
    assert success is False

    # Check with a hash misses a property and the schema does not have
    # a default value
    invalid_hash = Hash("stringProperty", "foo",
                        "boolProperty", True)
    success, value = sanitize_table_value(empty_binding,
                                          HashList([invalid_hash]))
    assert value == HashList([Hash("stringProperty", "foo",
                                   "uintProperty", 0,  # forced value
                                   "boolProperty", True)])
    assert success is False

    # Just one property, but no real validation
    valid_hash = Hash("stringProperty", "foobar")
    success, value = sanitize_table_value(readonly_binding,
                                          HashList([valid_hash]))
    assert value == HashList([Hash("stringProperty", "foobar")])
    assert success is True

    # Mixed table. We provide:
    # uintProperty with string, validation fails -> available default
    # missing string property, available default
    # missing bool property, no available default, forced
    valid_hash = Hash("uintProperty", "K")
    success, value = sanitize_table_value(mixed_default_binding,
                                          HashList([valid_hash]))
    assert value == HashList([Hash("stringProperty", "bar",
                                   "uintProperty", 20,
                                   "boolProperty", False)])
    assert success is False


def _assert_array(binding, value, dtype):
    array = validate_value(binding, value)
    np.testing.assert_array_equal(array, value)
    assert array.dtype == dtype


def test_default_value():
    """Test the default value generation of a binding"""
    binding = types.FloatBinding()
    # No default value provided, no `force` as default (False), value must
    # be `None` if not present
    value = get_default_value(binding)
    assert value is None
    # Force a default value for a float binding
    value = get_default_value(binding, force=True)
    assert value == 0.0
    # Apply a default value and retrieve default value
    binding.attributes[KARABO_SCHEMA_DEFAULT_VALUE] = 5.1
    value = get_default_value(binding, force=True)
    assert value == 5.1

    # Go through all other bindings
    binding = types.Int8Binding()
    value = get_default_value(binding, force=True)
    assert value == 0
    binding = types.Int16Binding()
    value = get_default_value(binding, force=True)
    assert value == 0
    binding = types.Int32Binding()
    value = get_default_value(binding, force=True)
    assert value == 0
    binding = types.Int64Binding()
    value = get_default_value(binding, force=True)
    assert value == 0
    binding = types.Uint8Binding()
    value = get_default_value(binding, force=True)
    assert value == 0
    binding = types.Uint16Binding()
    value = get_default_value(binding, force=True)
    assert value == 0
    binding = types.Uint32Binding()
    value = get_default_value(binding, force=True)
    assert value == 0
    binding = types.Uint64Binding()
    value = get_default_value(binding, force=True)
    assert value == 0

    binding = types.VectorHashBinding()
    value = get_default_value(binding, force=True)
    assert value == []

    binding = types.VectorInt8Binding()
    value = get_default_value(binding, force=True)
    assert value == []
    binding = types.VectorInt16Binding()
    value = get_default_value(binding, force=True)
    assert value == []
    binding = types.VectorInt32Binding()
    value = get_default_value(binding, force=True)
    assert value == []
    binding = types.VectorInt64Binding()
    value = get_default_value(binding, force=True)
    assert value == []
    binding = types.VectorUint8Binding()
    value = get_default_value(binding, force=True)
    assert value == []
    binding = types.VectorUint16Binding()
    value = get_default_value(binding, force=True)
    assert value == []
    binding = types.VectorUint32Binding()
    value = get_default_value(binding, force=True)
    assert value == []
    binding = types.VectorUint64Binding()
    value = get_default_value(binding, force=True)
    assert value == []

    binding = types.StringBinding()
    value = get_default_value(binding, force=True)
    assert value == ""

    binding = types.BoolBinding()
    value = get_default_value(binding, force=True)
    assert value is False

    binding = types.CharBinding()
    value = get_default_value(binding, force=True)
    assert value == ""

    binding = types.ByteArrayBinding()
    value = get_default_value(binding, force=True)
    assert value == bytearray([])


def test_default_value_minimum():
    """Test the default value generation of a binding"""
    attributes = {KARABO_SCHEMA_MIN_INC: 2}
    binding = types.Int8Binding(attributes=attributes)
    # 1. No default value provided, we force to minimum
    value = get_default_value(binding, force=True)
    assert value == 2
    # 2. Provide a default, higher than minimum
    attributes = {KARABO_SCHEMA_MIN_INC: 2, KARABO_SCHEMA_DEFAULT_VALUE: 5}
    binding = types.Int8Binding(attributes=attributes)
    value = get_default_value(binding, force=True)
    assert value == 5
    # 3. Check options
    attributes = {KARABO_SCHEMA_OPTIONS: [5, 6, 7]}
    binding = types.Int8Binding(attributes=attributes)
    value = get_default_value(binding, force=True)
    assert value == 5

    # ------------------------------

    attributes = {KARABO_SCHEMA_MIN_EXC: 0.0}
    binding = types.FloatBinding(attributes=attributes)
    # 1. No default value provided, we force to slighly above minimum
    value = get_default_value(binding, force=True)
    assert value > 0.0
    assert value < 1e-32
    # 2. Provide a default, higher than minimum
    attributes = {KARABO_SCHEMA_MIN_INC: 2.1, KARABO_SCHEMA_DEFAULT_VALUE: 5.2}
    binding = types.FloatBinding(attributes=attributes)
    value = get_default_value(binding, force=True)
    assert value == 5.2
    # 3. Check options
    attributes = {KARABO_SCHEMA_OPTIONS: [5.1, 6.2, 7.3]}
    binding = types.FloatBinding(attributes=attributes)
    value = get_default_value(binding, force=True)
    assert value == 5.1


def test_default_value_vector_minsize():
    """Test the default value generation of a binding"""
    attributes = {KARABO_SCHEMA_MIN_SIZE: 2}
    binding = types.VectorUint8Binding(attributes=attributes)
    # 1. No default value provided, we force to minimum with minSize 2
    value = get_default_value(binding, force=True)
    assert isinstance(value, np.ndarray)
    assert (value == [0, 0]).all()
    assert value.dtype == np.uint8

    # 2. Provide a default, higher than min size
    attributes = {KARABO_SCHEMA_MIN_SIZE: 2,
                  KARABO_SCHEMA_DEFAULT_VALUE: [3, 3, 3]}
    binding = types.Int8Binding(attributes=attributes)
    value = get_default_value(binding, force=True)
    assert value == [3, 3, 3]
    # Default value in this test provided without numpy type
    assert getattr(value, 'dtype', None) is None

    # 3. No default value provided, vector string of size 4
    attributes = {KARABO_SCHEMA_MIN_SIZE: 4}
    binding = types.VectorStringBinding(attributes=attributes)
    value = get_default_value(binding, force=True)
    assert value == ["", "", "", ""]


def test_validate_binding_configuration():
    schema = get_simple_props_schema()
    binding = build_binding(schema)

    # Note: validate binding configuration uses `validate_value`, which is
    # extensively tested. We check that the iteration works

    # 1. Leaf element testing
    config = Hash("boolProperty", False)
    fails = validate_binding_configuration(binding, config)
    # Boolean is fine
    assert fails.empty()

    config = Hash("intProperty", 2)
    fails = validate_binding_configuration(binding, config)
    assert fails.empty()

    config = Hash("intProperty", 2.0)
    fails = validate_binding_configuration(binding, config)
    assert "intProperty" in fails

    config = Hash("NOTINSCHEMA", False)
    # Right now, properties that are not in the schema, are not in fails
    # This might change in the future
    fails = validate_binding_configuration(binding, config)
    # Boolean is fine
    assert fails.empty()

    # 2. Node element testing
    # Test that elements in nodes are validated
    config = Hash("node.foo", "XFEL")
    fails = validate_binding_configuration(binding, config)
    assert "node.foo" not in fails

    config = Hash("node.bar", np.array([]), "node.charlie", 2)
    fails = validate_binding_configuration(binding, config)
    assert "node.bar" in fails
    assert "node.charlie" not in fails

    # 3. Table element testing
    # A valid config
    config = Hash("table", [Hash("foo", False, "bar", "XFEL", "charlie", 2),
                            Hash("foo", True, "bar", "laser", "charlie", 0)])
    fails = validate_binding_configuration(binding, config)
    assert "table" not in fails

    # Invalid, charlie becomes a float although integer
    config = Hash("table", [Hash("foo", False, "bar", "XFEL", "charlie", 3.0)])
    fails = validate_binding_configuration(binding, config)
    assert "table" in fails

    # Put None table
    config = Hash("table", None)
    fails = validate_binding_configuration(binding, config)
    assert "table" in fails


def test_convert_string():
    value, success = convert_string("2")
    assert value == 2
    assert success

    value, success = convert_string("00001")
    assert value == "00001"
    assert not success

    value, success = convert_string("1.3")
    assert value == 1.3
    assert success



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from numpy import uint64

from karabo.common.api import State
from karabo.common.scenemodel.api import SingleBitModel
from karabo.native import Configurable, Int32
from karabogui.indicators import STATE_COLORS
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..singlebit import SingleBit

COLORS = {s: str(STATE_COLORS[s]) for s in (State.ACTIVE, State.PASSIVE)}


class Object(Configurable):
    prop = Int32(defaultValue=0)


@pytest.fixture
def singlebit_setup(gui_app):
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    model = SingleBitModel()
    controller = SingleBit(proxy=proxy, model=model)
    controller.create(None)
    assert controller.widget is not None
    yield controller, proxy, model

    controller.destroy()
    assert controller.widget is None


def test_set_value(singlebit_setup):
    controller, proxy, model = singlebit_setup
    singlebit = controller._internal_widget

    model.invert = False
    model.bit = 5

    set_proxy_value(proxy, "prop", uint64(32))
    assert COLORS[State.ACTIVE] in singlebit.styleSheet()

    set_proxy_value(proxy, "prop", uint64(31))
    assert COLORS[State.PASSIVE] in singlebit.styleSheet()

    model.bit = 4
    assert COLORS[State.ACTIVE] in singlebit.styleSheet()

    model.invert = True
    assert COLORS[State.PASSIVE] in singlebit.styleSheet()


def test_read_only(singlebit_setup):
    controller, _, _ = singlebit_setup
    singlebit = controller._internal_widget

    controller.set_read_only(True)
    assert not singlebit.isEnabled()

    controller.set_read_only(False)
    assert singlebit.isEnabled()


def test_actions(singlebit_setup, mocker):
    controller, proxy, model = singlebit_setup
    actions = controller.widget.actions()
    invert_action = actions[0]
    change_action = actions[1]
    assert "invert" in invert_action.text().lower()
    assert "change" in change_action.text().lower()

    model.invert = False
    invert_action.trigger()
    assert model.invert

    sym = "karabogui.controllers.display.singlebit.QInputDialog"
    QInputDialog = mocker.patch(sym)
    QInputDialog.getInt.return_value = 3, True
    change_action.trigger()
    assert model.bit == 3

    proxy.binding.displayType = "bin|1,2,3"
    QInputDialog.getItem.return_value = "2:", True
    change_action.trigger()

    assert model.bit == 2



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa
from karabogui import icons, messagebox
from karabogui.binding.api import (
    REFERENCE_TYPENUM_TO_DTYPE, BaseBinding, BindingNamespace, BindingRoot,
    BoolBinding, ByteArrayBinding, CharBinding, FloatBinding, HashBinding,
    ImageBinding, Int8Binding, Int16Binding, Int32Binding, Int64Binding,
    IntBinding, NDArrayBinding, NodeBinding, NoneBinding,
    PipelineOutputBinding, ProxyStatus, SchemaBinding, SignedIntBinding,
    SlotBinding, StringBinding, TableBinding, Uint8Binding, Uint16Binding,
    Uint32Binding, Uint64Binding, UnsignedIntBinding, VectorBinding,
    VectorBoolBinding, VectorCharBinding, VectorDoubleBinding,
    VectorFloatBinding, VectorHashBinding, VectorInt8Binding,
    VectorInt16Binding, VectorInt32Binding, VectorInt64Binding,
    VectorNoneBinding, VectorNumberBinding, VectorStringBinding,
    VectorUint8Binding, VectorUint16Binding, VectorUint32Binding,
    VectorUint64Binding, WidgetNodeBinding, build_binding, convert_string,
    get_binding_array_value, get_binding_value, get_default_value,
    get_editor_value, get_min_max, get_min_max_size, get_native_min_max,
    has_min_max_attributes, is_signed_vector_integer,
    is_unsigned_vector_integer, is_vector_integer)
from karabogui.binding.proxy import DeviceProxy, PropertyProxy
from karabogui.const import WIDGET_MIN_HEIGHT
from karabogui.controllers.api import (
    DIMENSIONS, BaseArrayGraph, BaseBindingController, BaseLabelController,
    BaseLineEditController, BindingValidator, ListValidator, SimpleValidator,
    add_unit_label, axis_label, get_array_data, get_dimensions_and_encoding,
    get_image_data, has_options, is_proxy_allowed, register_binding_controller,
    with_display_type)
from karabogui.controllers.table.api import (
    BaseFilterTableController, BaseTableController, ComboBoxDelegate,
    KaraboTableView, TableButtonDelegate, TableModel, TableSortFilterModel,
    VectorButtonDelegate, VectorDelegate, is_state_display_type, list2string,
    string2list)
from karabogui.debug import profiler
from karabogui.dialogs.api import (
    FontDialog, FormatLabelDialog, ListEditDialog, TextDialog)
from karabogui.events import (
    KaraboEvent, broadcast_event, register_for_broadcasts,
    unregister_from_broadcasts)
from karabogui.fonts import get_font_metrics, get_font_size_from_dpi, get_qfont
from karabogui.graph.common.api import (
    AspectRatio, AxisItem, AxisType, BaseToolsetController, CrosshairROI,
    ExportTool, KaraboLegend, KaraboROI, KaraboViewBox, MouseMode, MouseTool,
    RectROI, ROITool, create_axis_items, create_button, create_tool_button,
    float_to_string, get_available_colors, get_default_brush, get_default_pen,
    get_pen_cycler, make_brush, make_pen, rgba_to_hex)
from karabogui.graph.common.const import DEFAULT_BAR_WIDTH, DEFAULT_PEN_WIDTH
from karabogui.graph.image.api import (
    SHOWN_AXES, AuxPlotItem, ColorBarWidget, ColorViewBox, KaraboImageItem,
    KaraboImageNode, KaraboImagePlot, KaraboImageView, KaraboImageViewBox,
    RollImage, bytescale, levels_almost_equal)
from karabogui.graph.plots.api import (
    KaraboPlotView, ScatterGraphPlot, TransformDialog, VectorBarGraphPlot,
    generate_baseline, generate_down_sample, get_view_range)
from karabogui.indicators import get_instance_info_icon, get_state_color
from karabogui.itemtypes import (
    ConfiguratorItemType, NavigationItemTypes, ProjectItemTypes)
from karabogui.logger import get_logger
from karabogui.request import (
    call_device_slot, get_macro_from_server, get_scene_from_server,
    onConfigurationUpdate, onSchemaUpdate, retrieve_default_scene,
    send_property_changes)
from karabogui.sceneview.api import get_proxy
from karabogui.singletons.api import (
    get_config, get_manager, get_network, get_panel_wrangler, get_topology)
from karabogui.topology.api import (
    SystemTopology, SystemTree, SystemTreeNode, get_macro_servers, getTopology,
    is_device_online, is_server_online)
from karabogui.util import (
    SignalBlocker, generateObjectName, get_reason_parts, get_spin_widget,
    getOpenFileName, getSaveFileName, show_wait_cursor, wait_cursor)
from karabogui.validators import (
    HexValidator, IntValidator, NumberValidator, RegexListValidator,
    RegexValidator)
from karabogui.widgets.api import (
    CodeBook, ElidingLabel, FrameWidget, Label, LineEdit, RangeSlider,
    SvgWidget, ToolBar)



#############################################################################
# Author: <alessandro.silenzi@xfel.eu>
# Created on October 26, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QAction, QProgressBar
from traits.api import Float, Instance, Tuple

from karabo.common.api import (
    KARABO_SCHEMA_MAX_EXC, KARABO_SCHEMA_MAX_INC, KARABO_SCHEMA_MIN_EXC,
    KARABO_SCHEMA_MIN_INC)
from karabo.common.scenemodel.api import DisplayProgressBarModel
from karabogui.binding.api import (
    FloatBinding, IntBinding, has_min_max_attributes)
from karabogui.const import WIDGET_MIN_HEIGHT
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)

NULL_RANGE = (0, 0)
PROGRESS_MAX = 200


def _orientation(is_vertical):
    return Qt.Vertical if is_vertical else Qt.Horizontal


def _scale(val, min_val, max_val):
    clamped = min(max_val, max(min_val, val))
    return int((clamped - min_val) / (max_val - min_val) * PROGRESS_MAX)


@register_binding_controller(ui_name='Progress Bar',
                             klassname='DisplayProgressBar',
                             binding_type=(FloatBinding, IntBinding),
                             is_compatible=has_min_max_attributes,
                             can_show_nothing=False)
class DisplayProgressBar(BaseBindingController):
    # The specific scene model class used by this widget
    model = Instance(DisplayProgressBarModel, args=())
    # Value scaling params
    _value_factors = Tuple(Float(-1), Float(-1))

    def create_widget(self, parent):
        widget = QProgressBar(parent)
        widget.setMinimumHeight(WIDGET_MIN_HEIGHT)
        orient_action = QAction("Change Orientation", widget)
        orient_action.triggered.connect(self._orientation_action)
        widget.addAction(orient_action)
        widget.setOrientation(_orientation(self.model.is_vertical))
        # Use a fixed integer range. Minimum defaults to 0.
        widget.setMaximum(PROGRESS_MAX)
        # update boolean context menu
        orient_action.setCheckable(True)
        orient_action.setChecked(self.model.is_vertical)

        return widget

    def binding_update(self, proxy):
        binding = proxy.binding
        self._eval_limits(binding)
        if self._value_factors == NULL_RANGE:
            msg = ('No proper configuration detected.\n Please define min and '
                   'max thresholds for {}'.format(proxy.key))
            self.widget.setToolTip(msg)

    def value_update(self, proxy):
        if self._value_factors == NULL_RANGE:
            return
        value = _scale(proxy.value, *self._value_factors)
        self.widget.setValue(value)

    def _eval_limits(self, binding):
        def _set_limits(low, high):
            """Avoid unset widget traits"""
            if self.widget is None:
                return
            self.widget.setMinimum(low)
            self.widget.setMaximum(high)

        attrs = binding.attributes
        min_val = attrs.get(KARABO_SCHEMA_MIN_INC,
                            attrs.get(KARABO_SCHEMA_MIN_EXC, 0))
        max_val = attrs.get(KARABO_SCHEMA_MAX_INC,
                            attrs.get(KARABO_SCHEMA_MAX_EXC, 0))

        self._value_factors = (min_val, max_val)
        if self._value_factors == NULL_RANGE:
            _set_limits(0, 0)
        else:
            _set_limits(0, PROGRESS_MAX)

    def _orientation_action(self):
        self._set_orientation(not self.model.is_vertical)

    def _set_orientation(self, is_vertical):
        self.model.is_vertical = is_vertical
        self.widget.setOrientation(_orientation(is_vertical))



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.

from qtpy.QtCore import Qt

from karabo.native import Hash
from karabogui.dialogs.api import GuiSessionInfo
from karabogui.testing import singletons, system_hash
from karabogui.topology.system_topology import SystemTopology


def test_session_dialog(gui_app, mocker):
    topology = SystemTopology()
    h = system_hash()
    devices = h["device"]
    devices.setElement("guidevice", Hash(), {"host": "BigIron",
                                             "classId": "GuiServerDevice",
                                             "capabilities": 0,
                                             "serverId": "karabo/GuiServer",
                                             "status": "ok"})
    topology.initialize(h)
    network = mocker.Mock()
    with singletons(topology=topology, network=network):
        # 1. Test init without deviceId as argument
        dialog = GuiSessionInfo()
        assert not dialog.isModal()
        assert dialog.list_widget.count() == 1

        item = dialog.list_widget.currentItem()
        assert item.data(Qt.DisplayRole) == "guidevice"
        network.onExecuteGeneric.assert_called_once()

        call_args = network.onExecuteGeneric.call_args
        args = call_args[0]
        assert args[0] == "guidevice"
        assert args[1] == "slotGetClientSessions"

        # Finish dialog!
        dialog.done(1)



from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import QDialog, QHBoxLayout, QToolButton, QWidget

from karabogui import icons
from karabogui.dialogs.api import ListEditDialog
from karabogui.widgets.edits import LineEditEditor

from .utils import string2list


class TableVectorEdit(QWidget):

    def __init__(self, binding, parent=None):
        super().__init__(parent)
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self.lineedit = LineEditEditor(parent)
        layout.addWidget(self.lineedit)
        self.setFocusProxy(self.lineedit)
        self.lineedit.setFocusPolicy(Qt.StrongFocus)
        self.list_edit = ListEditDialog(
            binding, duplicates_ok=True, parent=self)

        edit_button = QToolButton()
        edit_button.setText("Edit")
        edit_button.setIcon(icons.edit)
        edit_button.setMaximumSize(25, 25)
        edit_button.setFocusPolicy(Qt.NoFocus)
        edit_button.clicked.connect(self.on_open_dialog)
        layout.addWidget(edit_button)

    def setValidator(self, validator):
        self.lineedit.setValidator(validator)

    def setText(self, value):
        self.lineedit.setText(value)
        self.list_edit.set_list(string2list(value))

    def text(self):
        return self.lineedit.text()

    @Slot()
    def on_open_dialog(self):
        if self.list_edit.exec() == QDialog.Accepted:
            self.lineedit.setText(self.list_edit.string_values)
        self.lineedit.setFocus(Qt.PopupFocusReason)



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on February 10, 2012
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtWidgets import QAction, QDialog
from traits.api import Instance, String

from karabo.common.scenemodel.api import EditableSpinBoxModel
from karabogui.binding.api import (
    Int8Binding, Int16Binding, Int32Binding, Uint8Binding, Uint16Binding,
    Uint32Binding, get_editor_value, get_min_max)
from karabogui.controllers.api import (
    BaseBindingController, is_proxy_allowed, register_binding_controller)
from karabogui.dialogs.api import FormatLabelDialog
from karabogui.fonts import get_font_size_from_dpi
from karabogui.util import (
    MouseWheelEventBlocker, SignalBlocker, generateObjectName)
from karabogui.widgets.api import SpinBox

INT_BINDINGS = (Int8Binding, Int16Binding, Int32Binding, Uint8Binding,
                Uint16Binding, Uint32Binding)


@register_binding_controller(ui_name="Integer SpinBox", can_edit=True,
                             klassname="EditableSpinBox",
                             binding_type=INT_BINDINGS)
class EditableSpinBox(BaseBindingController):
    model = Instance(EditableSpinBoxModel, args=())
    _blocker = Instance(MouseWheelEventBlocker)
    _style_sheet = String

    def create_widget(self, parent):
        widget = SpinBox(parent)
        widget.valueChanged[int].connect(self._on_user_edit)
        widget.setSingleStep(1)
        self._blocker = MouseWheelEventBlocker(widget)
        widget.installEventFilter(self._blocker)

        objectName = generateObjectName(self)
        widget.setObjectName(objectName)

        self._style_sheet = (f"SpinBox#{objectName}" +
                             " {{ font: {}; font-size: {}pt; }}")

        # Formats
        format_action = QAction("Format field...", widget)
        format_action.triggered.connect(self._format_field)
        widget.addAction(format_action)

        self._apply_format(widget)

        return widget

    def binding_update(self, proxy):
        binding = proxy.binding
        label = None
        if binding is not None:
            label = f" {binding.unit_label}" if binding.unit_label else None
        low, high = get_min_max(binding)
        with SignalBlocker(self.widget):
            self.widget.setSuffix(label)
            self.widget.setRange(max(-0x80000000, low),
                                 min(0x7fffffff, high))

    def value_update(self, proxy):
        value = get_editor_value(proxy)
        if value is not None:
            with SignalBlocker(self.widget):
                self.widget.setValue(value)

    def state_update(self, proxy):
        enable = is_proxy_allowed(proxy)
        self.widget.setEnabled(enable)

    def _on_user_edit(self, value):
        if self.proxy.binding is None:
            return
        self.proxy.edit_value = value

    # -----------------------------------------------------------------------
    # Formatting methods

    def _format_field(self):
        dialog = FormatLabelDialog(font_size=self.model.font_size,
                                   font_weight=self.model.font_weight,
                                   parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            self.model.trait_set(font_size=dialog.font_size,
                                 font_weight=dialog.font_weight)
            self._apply_format()

    def _apply_format(self, widget=None):
        if widget is None:
            widget = self.widget

        # Apply font formatting
        sheet = self._style_sheet.format(
            self.model.font_weight,
            get_font_size_from_dpi(self.model.font_size))
        widget.setStyleSheet(sheet)
        widget.update()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os

from qtpy import uic
from qtpy.QtWidgets import QDialog

from karabogui.graph.common.const import TF_SCALING, TF_TRANSLATION


class ImageTransformsDialog(QDialog):

    def __init__(self, transforms, aspect_ratio, show_legend, parent=None):
        super().__init__(parent)
        self.setModal(False)
        # load ui file
        ui_path = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                               'transforms.ui')
        uic.loadUi(ui_path, self)

        # populate transform fields
        scale = transforms[TF_SCALING]
        offset = transforms[TF_TRANSLATION]

        self.ui_xscale.setValue(scale[0])
        self.ui_yscale.setValue(scale[1])
        self.ui_xoffset.setValue(offset[0])
        self.ui_yoffset.setValue(offset[1])

        # populate other fields
        self.ui_aspect_ratio.setCurrentIndex(aspect_ratio)
        self.ui_legend_checkbox.setChecked(show_legend)

        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

    @property
    def transforms(self):
        config = {
            "x_scale": self.ui_xscale.value(),
            "y_scale": self.ui_yscale.value(),
            "x_translate": self.ui_xoffset.value(),
            "y_translate": self.ui_yoffset.value(),
            "aspect_ratio": self.ui_aspect_ratio.currentIndex()
        }

        return config

    @property
    def show_scale(self):
        return self.ui_legend_checkbox.isChecked()

    @staticmethod
    def get(transforms, aspect_ratio, show_legend, parent=None):
        dialog = ImageTransformsDialog(transforms, aspect_ratio,
                                       show_legend, parent)
        result = dialog.exec() == QDialog.Accepted
        content = {}
        content.update(dialog.transforms)
        content["show_scale"] = dialog.show_scale

        return content, result



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QDialog

from karabo.common.project.macro import MacroModel
from karabogui.panels.macropanel import MacroPanel
from karabogui.singletons.configuration import Configuration
from karabogui.testing import development_system_hash, singletons
from karabogui.topology.api import SystemTopology


def test_save_dialog(gui_app, mocker):
    """Test the default file name on saving macro from panel"""

    # Name without slash
    model = MacroModel(simple_name="macro_foo")
    panel = MacroPanel(model=model)

    # To avoid actually writing the file mocking the 'open' function.
    mock_method = mocker.patch("builtins.open") # noqa

    mock_dialog = mocker.patch(
        'karabogui.panels.macropanel.getSaveFileName', return_value="foo.py")
    panel.on_save()
    assert mock_dialog.call_args.kwargs['selectFile'] == "macro_foo.py"

    # Name with slash
    panel.model.simple_name = "macro/foo"
    panel.on_save()
    assert mock_dialog.call_args.kwargs['selectFile'] == "macro_foo.py"


def test_on_debug_run(gui_app, mocker):
    """
    Test that 'on_debug_run' shows the 'DebugRunDialog' and runs the macro on
    specific server and set the configuration when the dialog is accepted.
    """
    model = MacroModel(simple_name="macro_foo")
    panel = MacroPanel(model=model)

    topology = SystemTopology()
    topology.initialize(development_system_hash())
    config = Configuration()

    mock_run = mocker.patch("karabogui.panels.macropanel.run_macro_debug")
    mock_dialog = mocker.patch("karabogui.panels.macropanel.DebugRunDialog")

    mock_dialog().exec.return_value = QDialog.Rejected
    with singletons(topology=topology, configuration=config):
        panel.on_debug_run()
        assert mock_run.call_count == 0
        mock_dialog().exec.return_value = QDialog.Accepted
        mock_dialog().comboBox.currentText.return_value = "devserver"
        panel.on_debug_run()
        mock_run.assert_called_once_with(
            model, serverId="devserver", parent=panel)
        assert config['macro_development'] == "devserver"


def test_repr(gui_app):
    model = MacroModel(simple_name="macro_foo")
    panel = MacroPanel(model=model)
    assert repr(panel) == "<MacroPanel macro=macro_foo>"



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from itertools import cycle
from weakref import WeakValueDictionary

from traits.api import Instance

from karabo.common.scenemodel.api import (
    VectorXYGraphModel, build_graph_config, build_model_config)
from karabogui import icons
from karabogui.binding.api import (
    VectorBoolBinding, VectorNumberBinding, get_binding_value)
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.graph.common.api import get_pen_cycler
from karabogui.graph.common.toolbar.widgets import create_button
from karabogui.graph.plots.api import (
    KaraboPlotView, generate_down_sample, get_view_range)
from karabogui.graph.plots.dialogs.data_analysis import DataAnalysisDialog


def _is_compatible(binding):
    """Don't allow plotting of boolean vectors"""
    return not isinstance(binding, VectorBoolBinding)


@register_binding_controller(ui_name='Vector XY Graph',
                             klassname='VectorXYGraph',
                             binding_type=VectorNumberBinding,
                             is_compatible=_is_compatible,
                             can_show_nothing=False)
class DisplayVectorXYGraph(BaseBindingController):
    """The vector xy plot graph class

    - First property proxy (dragged) will declare the x axis
    - Other property proxy (added) will declare the y axis curves
    """
    model = Instance(VectorXYGraphModel, args=())
    # Internal traits
    _curves = Instance(WeakValueDictionary, args=())
    _pens = Instance(cycle, allow_none=False)

    def create_widget(self, parent):
        widget = KaraboPlotView(parent=parent)
        widget.stateChanged.connect(self._change_model)
        widget.add_legend(visible=False)
        widget.add_cross_target()
        toolbar = widget.add_toolbar()
        widget.enable_data_toggle()
        widget.enable_export()

        widget.restore(build_model_config(self.model))

        button = create_button(
            icon=icons.data_analysis, checkable=False, tooltip="Data Analysis",
            on_clicked=self.show_data_analysis_dialog)
        toolbar.add_button(button)
        return widget

    def show_data_analysis_dialog(self):
        config = build_graph_config(self.model)
        data_analysis_dialog = DataAnalysisDialog(
            proxies=self.proxies[1:], config=config, baseline_proxy=self.proxy,
            parent=self.widget)
        data_analysis_dialog.show()

    def __pens_default(self):
        return get_pen_cycler()

    # ----------------------------------------------------------------

    def add_proxy(self, proxy):
        name = proxy.key
        curve = self.widget.add_curve_item(name=name, pen=next(self._pens))
        self._curves[proxy] = curve
        if len(self._curves) > 1:
            self.widget.set_legend(True)
        return True

    def value_update(self, proxy):
        value = get_binding_value(proxy.binding, [])
        if len(value) > 0:
            # The x-axis proxy changed!
            if proxy is self.proxy:
                for p, c in self._curves.items():
                    # since proxy is used as key for stored curves, before
                    # getting the previous values from the proxy, we have to
                    # check for Undefined
                    y_value = get_binding_value(p, [])
                    self._plot_data(c, value, y_value)
            else:
                c = self._curves.get(proxy, None)
                if c is None:
                    # Note: This can happen on start up ...
                    return

                x_value = get_binding_value(self.proxy, [])
                self._plot_data(c, x_value, value)

    def _plot_data(self, curve, x, y):
        """Plot the data x and y on the `curve`

        Take into account a size missmatch of the data
        """
        size_x = len(x)
        size_y = len(y)
        size = min(size_x, size_y)
        if size == 0:
            curve.setData([], [])
            return

        if size_x != size_y:
            x = x[:size]
            y = y[:size]

        rect = get_view_range(curve)
        x, y = generate_down_sample(y, x=x, rect=rect, deviation=False)
        curve.setData(x, y)

    # ----------------------------------------------------------------
    # Qt Slots

    def _change_model(self, content):
        self.model.trait_set(**content)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QToolBar
from traits.api import Dict, HasStrictTraits, Instance

from ..enums import MouseTool
from .factory import tool_factory
from .toolsets import get_toolset, register_toolset
from .widgets import ToolBar


class ToolbarController(HasStrictTraits):
    widget = Instance(QToolBar)

    toolsets = Dict
    buttons = Dict

    def __init__(self, parent=None, **traits):
        super().__init__(**traits)
        self.widget = ToolBar(parent=parent)

        default_tool_type = MouseTool
        controller = get_toolset(default_tool_type)
        self.toolsets[default_tool_type] = controller

    def set_background(self, color):
        """Set the background stylesheet of the toolbar widget"""
        self.widget.setBackground(color)

    def _toolsets_items_changed(self, event):
        """Add the buttons"""

        # Add separator in between toolsets
        if len(self.toolsets) > 1:
            self.widget.addSeparator()

        for controller in event.added.values():
            for tool, button in controller.buttons.items():
                self.widget.add_button(button)
                self.buttons[tool.name] = button

    def add_tool(self, tool):
        """Adds individual tools

        :param tool: An enumerable defining the tool
        """
        toolset = tool.__class__
        if toolset in self.toolsets:
            # We add the tool in the existing toolset
            controller = self.toolsets[toolset]
            button = controller.add(tool)
        else:
            # We add the tool as a separate button.
            button = tool_factory(tool)

        self.buttons[tool.name] = button
        self.widget.add_button(button)

        return button

    def add_button(self, button, key=None, separator=False):
        """Adds a custom button outside of the button factories"""
        # Add separator if specified
        if separator:
            self.widget.addSeparator()
        # Add button to widget
        self.widget.add_button(button)
        # Record if key is specified
        if key is not None:
            self.buttons[key] = button

    def add_toolset(self, tool_type, tools=None):
        """Adds a pre-configured set of tools which has interaction with each
           other."""
        if tool_type in self.toolsets:
            return

        controller = get_toolset(tool_type, tools=tools)
        if controller is not None:
            self.toolsets[tool_type] = controller
        return controller

    def register_toolset(self, toolset, klass):
        register_toolset(toolset, klass)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from collections import deque

from qtpy.QtWidgets import QAction, QInputDialog
from traits.api import Any, Instance, WeakRef

from karabo.common.scenemodel.api import ScatterGraphModel, build_model_config
from karabo.native import Timestamp
from karabogui import icons
from karabogui.binding.api import (
    BoolBinding, FloatBinding, IntBinding, PropertyProxy, get_binding_value)
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.graph.common.api import create_button
from karabogui.graph.plots.api import KaraboPlotView, ScatterGraphPlot

BUTTON_SIZE = (52, 32)
NUMERICAL_BINDINGS = (BoolBinding, FloatBinding, IntBinding)
MAX_NUM_POINTS = 1000

MIN_POINT_SIZE = 0.1
MAX_POINT_SIZE = 10.0


@register_binding_controller(ui_name="Scatter Graph", klassname="ScatterGraph",
                             binding_type=NUMERICAL_BINDINGS,
                             can_show_nothing=False)
class DisplayScatterGraph(BaseBindingController):
    """The scatter graph widget provides a scatter plot from property proxies

    - First property proxy (dragged) will declare the x axis
    - Second property proxy (added) will declare the y axis
    """
    model = Instance(ScatterGraphModel, args=())

    # Internal traits
    _y_proxy = Instance(PropertyProxy)
    _x_values = Instance(deque)
    _y_values = Instance(deque)
    _last_x_value = Any
    _plot = WeakRef(ScatterGraphPlot)

    _timestamp = Instance(Timestamp, args=())

    def create_widget(self, parent):

        widget = KaraboPlotView(parent=parent)
        widget.add_cross_target()
        toolbar = widget.add_toolbar()

        _btn_reset = create_button(checkable=False,
                                   icon=icons.reset,
                                   tooltip="Reset the plot",
                                   on_clicked=self._reset_plot)
        toolbar.add_button(_btn_reset)
        widget.stateChanged.connect(self._change_model)

        model = self.model
        self._x_values = deque(maxlen=model.maxlen)
        self._y_values = deque(maxlen=model.maxlen)

        self._plot = widget.add_scatter_item()

        # assigning proxy is safe and wanted here!
        self._last_x_value = get_binding_value(self.proxy)

        widget.restore(build_model_config(self.model))
        self._plot.setSize(self.model.psize)

        deque_action = QAction("Queue Size", widget)
        deque_action.triggered.connect(self._configure_deque)
        widget.addAction(deque_action)

        point_size_action = QAction("Point Size", widget)
        point_size_action.triggered.connect(self._configure_point_size)
        point_size_action.setIcon(icons.scatter)
        widget.addAction(point_size_action)

        return widget

    # ----------------------------------------------------------------

    def add_proxy(self, proxy):
        if self._y_proxy is None:
            self._y_proxy = proxy
            return True

        return False

    def remove_proxy(self, proxy):
        if self._y_proxy is not None:
            self._y_proxy = None
            self._reset_plot()
            return True

        return False

    def value_update(self, proxy):
        value = proxy.value
        if proxy is self.proxy:
            self._last_x_value = value
        elif proxy is self._y_proxy and self._last_x_value is not None:
            timestamp = proxy.binding.timestamp
            if self.widget is not None and timestamp != self._timestamp:
                self._timestamp = timestamp
                self._x_values.append(self._last_x_value)
                self._y_values.append(value)
                self._plot.setData(self._x_values, self._y_values)

    # ----------------------------------------------------------------
    # Qt Slots

    def _change_model(self, content):
        self.model.trait_set(**content)

    def _reset_plot(self):
        self._last_x_value = None
        self._x_values.clear()
        self._y_values.clear()
        self._plot.clear()

    def _configure_deque(self):
        maxlen, ok = QInputDialog.getInt(self.widget, "Number of Values",
                                         "Maxlen:", self.model.maxlen, 5,
                                         MAX_NUM_POINTS)
        if ok:
            self._last_x_value = None
            self._x_values, self._y_values = None, None
            self._x_values = deque(maxlen=maxlen)
            self._y_values = deque(maxlen=maxlen)
            self.model.maxlen = maxlen

    def _configure_point_size(self):
        psize, ok = QInputDialog.getDouble(self.widget, "Size of points",
                                           "Pointsize:", self.model.psize,
                                           MIN_POINT_SIZE, MAX_POINT_SIZE)
        if ok:
            self._plot.setSize(psize)
            self.model.psize = psize



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on September 12, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import QSize
from qtpy.QtWidgets import QStyledItemDelegate

from .utils import FIXED_ROW_HEIGHT


class ValueDelegate(QStyledItemDelegate):
    """A delegate for graceful size calculation for the value column

    This value delegate will prevent flickering of the column size
    """
    def sizeHint(self, option, index):
        return QSize(option.rect.width(), FIXED_ROW_HEIGHT)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.\
from abc import abstractmethod
from collections import namedtuple

from qtpy import uic
from qtpy.QtCore import QPointF, Qt, Slot
from qtpy.QtGui import QBrush, QColor, QFont, QIcon, QPen, QPixmap, QTransform
from qtpy.QtWidgets import (
    QColorDialog, QDialog, QGraphicsItem, QGraphicsPixmapItem)
from traits.api import ABCHasStrictTraits, Instance

from karabogui import icons
from karabogui.const import IS_MAC_SYSTEM
from karabogui.dialogs.font_dialog import FontDialog
from karabogui.dialogs.utils import get_dialog_ui
from karabogui.fonts import get_alias_from_font, get_qfont

BUTTON_SIZE = 10


class TextDialog(QDialog):
    """Simple Text dialog to allow user to enter the text and also to select
    the font and color for the text."""
    def __init__(self, font, color, parent=None):
        super().__init__(parent=parent)
        uic.loadUi(get_dialog_ui('simple_text_dialog.ui'), self)
        self.text_font = font
        self.text_color = color
        self.set_text_color_button()
        self.set_text_font_button()
        self.setModal(False)

    @property
    def text(self):
        return self.text_line_edit.text()

    @Slot()
    def on_pbFont_clicked(self):
        dialog = FontDialog(self.text_font, parent=self)
        if dialog.exec() == QDialog.Accepted:
            self.text_font = dialog.qfont
            self.set_text_font_button()

    @Slot()
    def on_pbTextColor_clicked(self):
        color = QColorDialog.getColor(initial=self.text_color, parent=self)
        if color.isValid():
            self.text_color = color
            self.set_text_color_button()

    def set_text_color_button(self):
        pixmap = QPixmap(24, 16)
        pixmap.fill(self.text_color)
        self.pbTextColor.setIcon(QIcon(pixmap))

    def set_text_font_button(self):
        """
        Update the  text and font family of the button. The font size
        should not be changed in order to keep the button size fixed.
        """
        qfont = QFont(self.text_font)
        qfont.setPointSize(BUTTON_SIZE)
        self.pbFont.setFont(qfont)
        family = get_alias_from_font(self.text_font.family())
        button_text = f"{family}, {self.text_font.pointSize()}pt"
        self.pbFont.setText(button_text)


class BaseDrawingTool(ABCHasStrictTraits):
    """Base class for the drawing tools in a QGraphicsScene"""

    # QGraphics Item to be drawn on the scene
    graphics_item = Instance(QGraphicsItem)

    # Starting point for the graphics item.
    start_pos = Instance(QPointF)

    # Annotation Color
    pen_color = Instance(QColor)

    @abstractmethod
    def mouse_down(self, scene, event):
        """Method to call on mouse press on the GraphicsScene"""

    @abstractmethod
    def mouse_move(self, scene, event):
        """Method to call on mouse move on the GraphicsScene"""

    @abstractmethod
    def mouse_up(self, scene, event):
        """Method to call on mouse move on the GraphicsScene."""

    def set_pen_color(self, color):
        self.pen_color = color


class LineTool(BaseDrawingTool):

    def mouse_down(self, scene, event):
        self.start_pos = event.scenePos()
        pen = QPen(QBrush(QColor(self.pen_color)), 2)
        self.graphics_item = scene.addLine(
            self.start_pos.x(), self.start_pos.y(),
            self.start_pos.x(), self.start_pos.y(), pen)

    def mouse_move(self, scene, event):
        pos = event.scenePos()
        if self.start_pos and self.graphics_item:
            self.graphics_item.setLine(
                self.start_pos.x(), self.start_pos.y(), pos.x(), pos.y())

    def mouse_up(self, scene, event):
        pass


class RectTool(BaseDrawingTool):

    def mouse_down(self, scene, event):
        pos = event.scenePos()
        pen = QPen(QBrush(QColor(self.pen_color)), 2)
        self.start_pos = pos
        self.graphics_item = scene.addRect(
            pos.x(), pos.y(), 0, 0, pen, QBrush())

    def mouse_move(self, scene, event):
        pos = event.scenePos()
        if self.start_pos and self.graphics_item:
            width = pos.x() - self.start_pos.x()
            height = pos.y() - self.start_pos.y()
            self.graphics_item.setRect(self.start_pos.x(), self.start_pos.y(),
                                       width, height)

    def mouse_up(self, scene, event):
        pass


class TextTool(BaseDrawingTool):
    """Allow to open a dialog to define the text to the QGraphicsScene"""

    color = Instance(QColor)
    font = Instance(QFont)

    def mouse_down(self, scene, event):
        pass

    def mouse_move(self, scene, event):
        pass

    def mouse_up(self, scene, event):
        if self.font is None:
            self.font = get_qfont()
        if self.color is None:
            self.color = QColor("black")
        parent = scene.parent()
        text_dialog = TextDialog(
            font=self.font,  color=self.color, parent=parent)
        if text_dialog.exec() == QDialog.Accepted:
            text = text_dialog.text.strip()
            font = text_dialog.text_font
            color = text_dialog.text_color
            if text:
                self.graphics_item = scene.addSimpleText(text, font)
                brush = QBrush(color)
                self.graphics_item.setBrush(brush)
                self.graphics_item.setPos(event.scenePos())
            self.color = color
            self.font = font
        # To avoid LogBookPreview dialog hiding behind the main window
        if IS_MAC_SYSTEM:
            parent.raise_()


class EraserTool(BaseDrawingTool):

    def mouse_down(self, scene, event):
        if event.button() == Qt.LeftButton:
            item = scene.itemAt(event.scenePos(), QTransform())
            if item and not isinstance(item, QGraphicsPixmapItem):
                scene.removeItem(item)

    def mouse_move(self, scene, event):
        pass

    def mouse_up(self, scene, event):
        pass


class CropTool(BaseDrawingTool):

    def mouse_down(self, scene, event):
        pos = event.scenePos()
        pen = QPen(QBrush(QColor(0, 0, 255, 127)), 2)
        self.start_pos = pos
        self.graphics_item = scene.addRect(
            pos.x(), pos.y(), 0, 0, pen, QBrush())

    def mouse_move(self, scene, event):
        pos = event.scenePos()
        if self.start_pos and self.graphics_item:
            width = pos.x() - self.start_pos.x()
            height = pos.y() - self.start_pos.y()
            self.graphics_item.setRect(self.start_pos.x(), self.start_pos.y(),
                                       width, height)
            self.graphics_item.setBrush(QColor(0, 0, 255, 127))

    def mouse_up(self, scene, event):
        scene.removeItem(self.graphics_item)
        rect = self.graphics_item.rect()
        scene.crop(rect)
        # Remove the remnants of annotations.
        for item in scene.items():
            if item and not isinstance(item, QGraphicsPixmapItem):
                scene.removeItem(item)


TOOLS_FACTORY = namedtuple("TOOL_FACTORY", ["tooltip", "icon", "drawing_tool"])


def get_tools():
    return [
        TOOLS_FACTORY(tooltip="Draw Line", icon=icons.line,
                      drawing_tool=LineTool),
        TOOLS_FACTORY(tooltip="Draw Rectangle", icon=icons.rect,
                      drawing_tool=RectTool),
        TOOLS_FACTORY(tooltip="Add Text to the Image", icon=icons.text,
                      drawing_tool=TextTool),
        TOOLS_FACTORY(tooltip="Delete item", icon=icons.eraser,
                      drawing_tool=EraserTool),
        TOOLS_FACTORY(tooltip="Crop image", icon=icons.crop,
                      drawing_tool=CropTool),
             ]



#############################################################################
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from traits.api import Instance

from karabo.common.api import KARABO_SCHEMA_ROW_SCHEMA
from karabo.common.scenemodel.api import (
    FilterTableElementModel, TableElementModel)
from karabogui.binding.api import VectorHashBinding
from karabogui.controllers.api import register_binding_controller
from karabogui.controllers.table.api import (
    BaseFilterTableController, BaseTableController)


def _is_compatible(binding):
    return KARABO_SCHEMA_ROW_SCHEMA in binding.attributes


@register_binding_controller(ui_name="Table Element",
                             klassname="EditableTableElement", can_edit=True,
                             binding_type=VectorHashBinding, priority=100,
                             is_compatible=_is_compatible)
class EditableTableElement(BaseTableController):
    """The editable version of the table element"""
    model = Instance(TableElementModel, args=())


@register_binding_controller(ui_name="Display Table Element",
                             klassname="DisplayTableElement", priority=90,
                             binding_type=VectorHashBinding,
                             is_compatible=_is_compatible)
class DisplayTableElement(BaseTableController):
    """The display version of the table element"""
    model = Instance(TableElementModel, args=())


@register_binding_controller(ui_name="Editable Filter Table Element",
                             klassname="EditableFilterTableElement",
                             can_edit=True,
                             binding_type=VectorHashBinding, priority=80,
                             is_compatible=_is_compatible)
class EditableFilterTableElement(BaseFilterTableController):
    """The editable version of the filter table element"""
    model = Instance(FilterTableElementModel, args=())


@register_binding_controller(ui_name="Display Filter Table Element",
                             klassname="DisplayFilterTableElement",
                             priority=80,
                             binding_type=VectorHashBinding,
                             is_compatible=_is_compatible)
class DisplayFilterTableElement(BaseFilterTableController):
    """The display version of the filter table element"""
    model = Instance(FilterTableElementModel, args=())



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import uuid

import pytest

from karabo.common.api import KARABO_PROJECT_MANAGER
from karabo.common.project.api import ProjectModel
from karabo.native import Hash
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.singletons.configuration import Configuration
from karabogui.singletons.db_connection import ProjectDatabaseConnection
from karabogui.singletons.mediator import Mediator
from karabogui.testing import singletons
from karabogui.util import process_qt_events


def test_connection(gui_app, mocker):
    mediator = Mediator()

    # Configure the default domain
    config = Configuration()
    domain = f"KaraboGUI-{uuid.uuid4()}"
    config["domain"] = domain

    network = mocker.Mock()
    with singletons(network=network, configuration=config,
                    mediator=mediator):
        assert len(mediator._listeners.keys()) == 0
        db_conn = ProjectDatabaseConnection()
        # db connection registers handlers in the mediator
        assert len(mediator._listeners.keys()) == 2
        assert db_conn.default_domain == domain

        # 1. Get the available domains
        db_conn.get_available_domains()
        network.onListProjectDomains.assert_called_once()

        # 2. Get available projects in domain
        db_conn.get_available_project_data(domain, "project")
        network.onProjectListItems.assert_called_with(
            KARABO_PROJECT_MANAGER, domain, 'project')

        # 3. Get projects
        db_conn.get_projects_with_device(domain, "karaboDevice")
        network.onProjectListProjectsWithDevice.assert_called_with(
            KARABO_PROJECT_MANAGER, domain, "karaboDevice"
        )

        # 4. Read data from project db
        assert db_conn.is_reading() is False
        assert db_conn.is_writing() is False
        assert db_conn.is_processing() is False

        project_uuid = str(uuid.uuid4())
        model = ProjectModel(uuid=project_uuid)
        model.initialized = True

        db_conn.retrieve(domain, project_uuid, model)
        assert db_conn.is_writing() is False
        assert db_conn.is_reading() is True
        assert db_conn.is_processing() is True

        # Network did not send yet
        network.onProjectLoadItems.assert_not_called()
        db_conn.flush()
        network.onProjectLoadItems.assert_called_with(
            KARABO_PROJECT_MANAGER,
            [Hash("domain", domain, "uuid", project_uuid,
                  "item_type", "project")]
        )

        data = {"success": True,
                "items": [Hash("domain", domain, "uuid", project_uuid,
                               "date", "", "xml", "")]}
        broadcast_event(KaraboEvent.ProjectItemsLoaded, data)
        process_qt_events(timeout=10)
        assert db_conn.is_reading() is False
        assert db_conn.is_processing() is False

        # And the `False` success case
        db_conn.retrieve(domain, project_uuid, model)
        data = {"success": False,
                "items": [Hash("domain", domain, "uuid", project_uuid,
                               "xml", "", "date", "")]}
        broadcast_event(KaraboEvent.ProjectItemsLoaded, data)
        process_qt_events(timeout=10)
        assert db_conn.is_reading() is False
        assert db_conn.is_processing() is False

        # 5. Save data in project db
        db_conn.store(domain, project_uuid, model)
        assert db_conn.is_writing() is True
        assert db_conn.is_reading() is False
        assert db_conn.is_processing() is True
        network.onProjectSaveItems.assert_not_called()

        db_conn.flush()
        network.onProjectSaveItems.assert_called_once()
        data = {"success": "True",
                "items": [Hash("domain", domain, "uuid", project_uuid,
                               "date", "", "success", True,
                               "xml", "")]}

        broadcast_event(KaraboEvent.ProjectItemsSaved, data)
        process_qt_events(timeout=10)
        assert db_conn.is_writing() is False
        assert db_conn.is_processing() is False

        # 6 Save a failed item top level success
        db_conn.store(domain, project_uuid, model)
        db_conn.flush()
        data = {"success": False,
                "items": [Hash("domain", domain, "uuid", project_uuid,
                               "date", "", "success", False,
                               "xml", "")]}
        broadcast_event(KaraboEvent.ProjectItemsSaved, data)
        process_qt_events(timeout=10)
        assert db_conn.is_processing() is False

        # 7. Save an item, low level success failure
        db_conn.store(domain, project_uuid, model)
        db_conn.flush()
        assert db_conn.is_processing() is True
        data = {"success": True,
                "items": [Hash("domain", domain, "uuid", project_uuid,
                               "date", "", "success", False,
                               "xml", "", "reason", "Conflict")]}
        mbox_path = "karabogui.singletons.db_connection.messagebox"
        mb = mocker.patch(mbox_path)
        broadcast_event(KaraboEvent.ProjectItemsSaved, data)
        process_qt_events(timeout=10)
        assert db_conn.is_processing() is False
        mb.show_error.assert_called_with("Conflict")

        # 8. Update project attribute
        db_conn.update_trashed(domain, "project", project_uuid, True)

        item = Hash('domain', domain, 'uuid', project_uuid, 'value', True,
                    'item_type', "project")
        network.onProjectUpdateTrashed.assert_called_with(
            KARABO_PROJECT_MANAGER, items=item)

        # 9. Configuration
        db_conn.default_domain = "New"
        assert config["domain"] == "New"

        # 10. Database scene
        scene_uuid = str(uuid.uuid4())
        token = db_conn.get_database_scene("scene", scene_uuid, 1)
        network.onExecuteGeneric.assert_called_with(
            KARABO_PROJECT_MANAGER, 'slotGetScene',
            Hash(
                "name", "scene",
                "uuid", scene_uuid,
                "domain", "New"), token=token)


if __name__ == "__main__":
    pytest.main()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on June 9, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os.path as op
from contextlib import contextmanager
from unittest import mock

import pytest
from qtpy.QtCore import QPoint
from qtpy.QtGui import QRegion
from qtpy.QtWidgets import QBoxLayout, QDialog

import karabo.common.scenemodel.api as sm
import karabo.common.scenemodel.tests as sm_tests
from karabo.common.api import set_initialized_flag, set_modified_flag
from karabo.common.scenemodel.tests.utils import single_model_round_trip
from karabogui.binding.api import (
    DeviceProxy, ImageBinding, Int8Binding, IntBinding, NDArrayBinding,
    NodeBinding, PipelineOutputBinding, PropertyProxy, ProxyStatus,
    SignedIntBinding, TableBinding, Uint8Binding, UnsignedIntBinding,
    VectorBinding, VectorHashBinding, VectorNoneBinding, VectorNumberBinding,
    VectorUint8Binding, WidgetNodeBinding, build_binding)
from karabogui.const import IS_MAC_SYSTEM
from karabogui.controllers.registry import get_model_controller
from karabogui.testing import (
    ALL_PROPERTIES_MAP, assert_no_throw, get_all_props_schema)

from ..api import SceneView
from ..layout.api import GroupLayout
from ..tools.clipboard import (
    SceneCopyAction, SceneCutAction, ScenePasteAction, ScenePasteReplaceAction)
from ..utils import add_offset, calc_relative_pos

DATA_DIR = op.join(op.abspath(op.dirname(sm_tests.__file__)), 'data')
INKSCAPE_DIR = op.join(DATA_DIR, 'inkscape')
UNKNOWN_XML_MODEL = sm.UnknownXMLDataModel(
    tag='circle',
    attributes={'cx': '100', 'cy': '100', 'r': '100'}
)


def load_file(view, dir_, file_name):
    assert view.scene_model is None
    assert isinstance(view.layout, GroupLayout)

    scene_model = sm.read_scene(op.join(dir_, file_name))
    scene_model.children.append(UNKNOWN_XML_MODEL)
    view.update_model(scene_model)
    assert isinstance(view.scene_model, sm.SceneModel)
    assert view.scene_model.width == view.width()
    assert view.scene_model.height == view.height()


def test_loading_karabo_svg(gui_app):
    view = SceneView()
    load_file(view, DATA_DIR, "all.svg")


def test_loading_inkscape_svg(gui_app):
    view = SceneView()
    load_file(view, INKSCAPE_DIR, "shapes.svg")


OFFSET = (10, 20)
GET_PROXY_PATH = "karabogui.sceneview.widget.container.get_proxy"
DEVICE_NAME = "Device"
ALL_PROPS_BINDINGS = {binding: name
                      for name, binding in ALL_PROPERTIES_MAP.items()}
REIMPLEMENTED_BINDINGS = {
    ImageBinding: NodeBinding,
    IntBinding: Int8Binding,
    NDArrayBinding: NodeBinding,
    PipelineOutputBinding: NodeBinding,
    SignedIntBinding: Int8Binding,
    TableBinding: VectorHashBinding,
    UnsignedIntBinding: Uint8Binding,
    VectorNumberBinding: VectorUint8Binding,
    VectorNoneBinding: VectorBinding,
    WidgetNodeBinding: NodeBinding}
SCENE_VIEW_PATH = "karabogui.sceneview.view.SceneView"
MAP_FROM_GLOBAL_PATH = SCENE_VIEW_PATH + ".mapFromGlobal"
VISIBLE_REGION_PATH = SCENE_VIEW_PATH + ".visibleRegion"
REPLACE_DIALOG_PATH = "karabogui.sceneview.tools.clipboard.ReplaceDialog"


# -----------------------------------------------------------------------------
# Fixtures:

@pytest.fixture
def setup_scene_view(gui_app):
    # Setup device
    schema = get_all_props_schema()
    binding = build_binding(schema)
    device = DeviceProxy(binding=binding,
                         server_id='Fake',
                         device_id=DEVICE_NAME,
                         status=ProxyStatus.OFFLINE)

    proxy_map = {name: PropertyProxy(root_proxy=device, path=name)
                 for name in ALL_PROPS_BINDINGS.values()}

    # Setup scene view
    scene_view = SceneView()
    scene_model = sm.SceneModel()
    yield proxy_map, scene_view, scene_model
    scene_view.destroy()


@pytest.fixture
def setup_load_scene_model(setup_scene_view):
    proxy_map, scene_view, _ = setup_scene_view
    yield proxy_map, scene_view
    # additional teardown
    with assert_no_throw():
        scene_view._update_widget_states()


# -----------------------------------------------------------------------------
# Assertions:

def assert_scene_element(scene_view, proxy_map, actual, expected,
                         modified=None):
    assert_element(scene_view, actual, expected, modified=modified)

    # Save and reread the model
    read_model = single_model_round_trip(actual)
    set_models_to_scene(scene_view, proxy_map, read_model)
    assert_element(scene_view, read_model, expected, modified=False)

    # Revert back the original model
    set_models_to_scene(scene_view, proxy_map, actual)


def assert_element(scene_view, actual, expected, modified=False):
    # Check geom traits
    geom_traits = ["x", "y", "width", "height"]
    actual_values = actual.trait_get(geom_traits)
    expected_values = expected.trait_get(geom_traits)
    assert actual_values == expected_values

    # Check if model is modified
    if modified is None:
        modified = expected.modified
    assert actual.modified == modified

    # Check widget
    widget = scene_view._scene_obj_cache.get(actual)
    scene_geom = widget.geometry()
    model_geom = actual.trait_get("x", "y", "width", "height")
    assert scene_geom.x() == model_geom.get("x", 0)
    assert scene_geom.y() == model_geom.get("y", 0)
    assert scene_geom.width() == model_geom.get("width")
    assert scene_geom.height() == model_geom.get("height")

    # Iterate over children
    actual_traits = actual.trait_names()
    expected_traits = expected.trait_names()
    if "children" in actual_traits and "children" in expected_traits:
        iter_children = zip(actual.children, expected.children)
        for act_child, exp_child in iter_children:
            assert_element(scene_view, act_child, exp_child, modified=modified)


def assert_geometry(scene_view, proxy_map, model_klass, **traits):

    def assert_single_model(model):
        model.trait_set(x=10, y=10, width=100, height=30)
        assign_key(model)
        set_initialized_flag(model, value=True)
        expected = copy_model(model)

        set_models_to_scene(scene_view, proxy_map, model)
        assert_scene_element(scene_view, proxy_map, model, expected)

        # Translate the widget
        apply_changes(model, expected)
        assert_scene_element(scene_view, proxy_map, model, expected,
                             modified=True)

    def assert_valid_layout(model):
        actual = get_layout_model(model)
        expected = copy_model(actual)

        set_models_to_scene(scene_view, proxy_map, actual)
        assert_scene_element(scene_view, proxy_map, actual, expected,
                             modified=False)

        # Translate the widget
        apply_changes(actual, expected)
        assert_scene_element(scene_view, proxy_map, actual, expected,
                             modified=True)

    def apply_changes(actual, expected):
        # Set the offset of the actual model
        widget = scene_view._scene_obj_cache.get(actual)
        widget.translate(QPoint(*OFFSET))

        expected.x += OFFSET[0]
        expected.y += OFFSET[1]
        if "children" in expected.trait_names():
            for child in expected.children:
                child.x += OFFSET[0]
                child.y += OFFSET[1]

    assertions = [assert_single_model, assert_valid_layout]
    for assertion in assertions:
        assertion(model_klass(**traits))
    # Dead layouts mean that the children widgets do not follow the desired
    # geometry. We let Qt handle such cases, which is as such on the
    # previous implementations. There's no need to check this.


def assert_clipboard_action(proxy_map, scene_view, *models):
    pos = (30, 30)
    # cut and paste
    with setup_and_test_selection(proxy_map, scene_view, *models,
                                  pos=pos) as scene_model:
        SceneCutAction().perform(scene_view)
        with setup_paste_action(scene_view, pos):
            ScenePasteAction().perform(scene_view)
    assert_cut_models(scene_view, scene_model, *models)

    # cut and paste with replacing
    with setup_and_test_selection(proxy_map, scene_view, *models,
                                  pos=pos) as scene_model:
        SceneCutAction().perform(scene_view)
        pastereplace_models(scene_view, *models, pos=pos)
    assert_cut_models(scene_view, scene_model, *models)

    # copy and paste
    with setup_and_test_selection(proxy_map, scene_view, *models,
                                  pos=pos) as scene_model:
        SceneCopyAction().perform(scene_view)
        with setup_paste_action(scene_view, pos):
            ScenePasteAction().perform(scene_view)
    assert_copied_models(scene_view, scene_model, *models)

    # copy and paste with replacing
    with setup_and_test_selection(proxy_map, scene_view, *models,
                                  pos=pos) as scene_model:
        SceneCopyAction().perform(scene_view)
        pastereplace_models(scene_view, *models, pos=pos)
    assert_copied_models(scene_view, scene_model, *models)


def assert_cut_models(scene_view, scene_model, *models):
    copied_models = [o.model for o in scene_view.selection_model]
    assert len(scene_model.children) == len(models)
    # Check if the copied models are in the scene
    for copied in copied_models:
        assert copied in scene_model.children
    # Check if the models are deleted in the scene
    for model in models:
        assert model not in scene_model.children


def assert_copied_models(scene_view, scene_model, *models):
    copied_models = [o.model for o in scene_view.selection_model]
    assert len(scene_model.children) == len(models) * 2
    for model in copied_models + list(models):
        assert model in scene_model.children


def assert_model(actual, expected, offset=(0, 0)):
    trait_names = expected.copyable_trait_names()

    # Recurse over the children models
    if "children" in trait_names:
        trait_names.remove("children")
        iter_children = zip(actual.children, expected.children)
        for actual_child, expected_child in iter_children:
            assert_model(actual_child, expected_child, offset=offset)

    # Check children models that are not linked as children (e.g., marker)
    if "marker" in trait_names:
        trait_names.remove("marker")
        assert_model(actual.marker, expected.marker)

    # Check if IDs are not the same since we generate a new ID for the
    # pasted models
    if "id" in trait_names:
        trait_names.remove("id")
        is_valid = actual.id != expected.id
        # If actual.id is blank, the valid is for the expected.id
        # to be blank also.
        if not actual.id:
            is_valid = not is_valid
        assert is_valid

    # Correct expected position
    copied = copy_model(expected)
    add_offset(copied, x=offset[0], y=offset[1])

    # Check if the rest of the traits are equivalent
    assert type(actual), type(expected)
    assert actual.trait_get(trait_names) == copied.trait_get(trait_names)


# -----------------------------------------------------------------------------
# Helper functions:

def assign_key(model):
    controller = get_model_controller(model)
    if controller is None:
        return None
    # Use a compatible binding
    ctrait = controller.class_traits().get("_binding_type")
    binding = get_binding(ctrait.default_value()[1])
    model.keys = [f"{DEVICE_NAME}.{ALL_PROPS_BINDINGS[binding]}"]


def copy_model(model):
    copied = model.clone_traits(copy="deep")
    set_initialized_flag(copied, value=model.initialized)
    set_modified_flag(copied, value=model.modified)
    return copied


def set_models_to_scene(scene_view, proxy_map, *models):

    def get_proxy(_, name):
        """Return a `PropertyProxy` instance for a given device and
        property path. This mocks the get_proxy called by
        ControllerContainer."""
        return proxy_map.get(name)

    # Check if there are defs in the models
    scene_model = sm.SceneModel(children=list(models))
    with mock.patch(GET_PROXY_PATH, new=get_proxy):
        scene_view.update_model(scene_model)
        scene_view.update()
    return scene_model


def get_layout_model(model, x=10, y=10, modified=False):
    # Initialize children models
    height = 30
    label_model = sm.LabelModel(width=100, height=height, text="bar")
    display_model = sm.DisplayLabelModel(width=200, height=height)
    model.trait_set(width=300, height=height)
    children_model = [label_model, display_model, model]

    # Initialize layout model
    hbox_model = sm.BoxLayoutModel(x=x, y=y, children=children_model,
                                   direction=QBoxLayout.LeftToRight)

    # Correct the position from the initial position and width
    x0, y0 = (x, y)
    for child in children_model:
        assign_key(child)
        child.trait_set(x=x0, y=y0)
        # Add child width for the next iteration
        x0 += child.width
    hbox_model.height = height
    hbox_model.width = sum([child.width for child in children_model])
    set_initialized_flag(hbox_model, value=True)
    set_modified_flag(hbox_model, value=modified)
    return hbox_model


def get_binding(bindings):
    for binding in bindings:
        if binding in ALL_PROPS_BINDINGS:
            return binding

        binding = REIMPLEMENTED_BINDINGS.get(binding)

        # We found a compatible binding
        if binding in ALL_PROPS_BINDINGS:
            return binding


def pastereplace_models(scene_view, *models, device_id="NewDevice",
                        pos=(0, 0)):
    mapped_device_ids = {DEVICE_NAME: device_id}

    # patch replace dialog
    exec_patch = mock.patch(REPLACE_DIALOG_PATH + ".exec",
                            return_value=QDialog.Accepted)
    mapped_patch = mock.patch(REPLACE_DIALOG_PATH + ".mappedDevices",
                              return_value=mapped_device_ids)
    with setup_paste_action(scene_view, pos), exec_patch, mapped_patch:
        action = ScenePasteReplaceAction()
        action.perform(scene_view)
        # Correct the models for checking later
        for m in models:
            action._set_widget_model_keys(m, mapped_device_ids)


def get_label_model(text, x=0, y=0):
    FONT_SPEC = 'Source Sans Pro,48,-1,5,50,0,0,0,0,0'
    traits = {'x': x, 'y': y, 'height': 100, 'width': 100,
              'text': text, 'font': FONT_SPEC,
              'foreground': '#000000', 'background': '#ffffff',
              'frame_width': 0}
    return sm.LabelModel(**traits)


@contextmanager
def setup_and_test_selection(proxy_map, scene_view, *models, pos=(0, 0)):
    try:
        # Reset scene view
        scene_model = set_models_to_scene(scene_view, proxy_map)
        assert len(scene_model.children) == 0

        # Add models to scene
        scene_model = set_models_to_scene(scene_view, proxy_map, *models)
        assert len(scene_model.children) == len(models)
        # Copy-paste
        scene_view.select_models(*models)
        yield scene_model
    finally:
        # Checked copied model has same traits as read model
        copied_models = [o.model for o in scene_view.selection_model]
        x0, y0 = calc_relative_pos(models)
        offset = (pos[0] - x0, pos[1] - y0)
        for copied_model, read_model in zip(copied_models, models):
            assert_model(copied_model, read_model, offset=offset)


@contextmanager
def setup_paste_action(scene_view, pos=(0, 0)):
    try:
        # patch cursor method
        map_from_global_patch = mock.patch(MAP_FROM_GLOBAL_PATH,
                                           return_value=QPoint(*pos))
        # patch visible region
        region = QRegion(0, 0, scene_view.width(), scene_view.height())
        visible_region_patch = mock.patch(VISIBLE_REGION_PATH,
                                          return_value=region)
        with map_from_global_patch, visible_region_patch:
            yield
    finally:
        pass


# -----------------------------------------------------------------------------
# Scene view tests:

def test_scene_view_basics(setup_scene_view):
    _, scene_view, scene_model = setup_scene_view
    scene_view.update_model(scene_model)
    scene_geom = scene_view.geometry()
    assert scene_geom.width() == scene_model.width
    assert scene_geom.height() == scene_model.height


# -----------------------------------------------------------------------------
# Scene load model tests:

def test_display_widgets(setup_load_scene_model):
    """Testing the loading of display widgets"""
    proxy_map, scene_view = setup_load_scene_model
    assert_geometry(scene_view, proxy_map, sm.CheckBoxModel)
    assert_geometry(scene_view, proxy_map, sm.ColorBoolModel)
    assert_geometry(scene_view, proxy_map, sm.DisplayCommandModel)
    assert_geometry(scene_view, proxy_map, sm.DisplayIconCommandModel)
    assert_geometry(scene_view, proxy_map, sm.DisplayLabelModel)
    assert_geometry(scene_view, proxy_map, sm.DisplayListModel)
    assert_geometry(scene_view, proxy_map, sm.DisplayProgressBarModel)
    assert_geometry(scene_view, proxy_map, sm.DisplayStateColorModel)
    assert_geometry(scene_view, proxy_map, sm.DisplayTextLogModel)
    assert_geometry(scene_view, proxy_map, sm.DisplayTimeModel)
    assert_geometry(scene_view, proxy_map, sm.ErrorBoolModel)
    assert_geometry(scene_view, proxy_map, sm.EvaluatorModel)
    assert_geometry(scene_view, proxy_map, sm.GlobalAlarmModel)
    assert_geometry(scene_view, proxy_map, sm.HexadecimalModel)
    assert_geometry(scene_view, proxy_map, sm.LampModel)
    assert_geometry(scene_view, proxy_map, sm.LineEditModel)
    assert_geometry(scene_view, proxy_map, sm.SingleBitModel)
    assert_geometry(scene_view, proxy_map, sm.TableElementModel)
    assert_geometry(scene_view, proxy_map, sm.WidgetNodeModel)


def test_editable_widgets(setup_load_scene_model):
    """Testing the geometry loading of editable widgets"""
    proxy_map, scene_view = setup_load_scene_model
    assert_geometry(scene_view, proxy_map, sm.CheckBoxModel,
                    klass="EditableCheckBox")
    assert_geometry(scene_view, proxy_map, sm.EditableChoiceElementModel)
    assert_geometry(scene_view, proxy_map, sm.EditableComboBoxModel)
    assert_geometry(scene_view, proxy_map, sm.DoubleLineEditModel)
    assert_geometry(scene_view, proxy_map, sm.EditableListElementModel)
    assert_geometry(scene_view, proxy_map, sm.EditableListModel)
    assert_geometry(scene_view, proxy_map, sm.EditableSpinBoxModel)
    assert_geometry(scene_view, proxy_map, sm.FloatSpinBoxModel)
    assert_geometry(scene_view, proxy_map, sm.IntLineEditModel)
    assert_geometry(scene_view, proxy_map, sm.LineEditModel,
                    klass="EditableLineEdit")
    assert_geometry(scene_view, proxy_map, sm.TableElementModel,
                    klass="EditableTableElement")
    assert_geometry(scene_view, proxy_map, sm.TickSliderModel)


def test_icon_widgets(setup_load_scene_model):
    proxy_map, scene_view = setup_load_scene_model
    assert_geometry(scene_view, proxy_map, sm.DigitIconsModel)
    # Note: IconSet is deprecated and now becomes a ImageRenderer
    # assert_geometry(scene_view, proxy_map, sm.DisplayIconsetModel,
    #                 data=b"<svg></svg>")
    assert_geometry(scene_view, proxy_map, sm.SelectionIconsModel)
    assert_geometry(scene_view, proxy_map, sm.TextIconsModel)


def test_image_graph_widgets(setup_load_scene_model):
    proxy_map, scene_view = setup_load_scene_model
    assert_geometry(scene_view, proxy_map, sm.DetectorGraphModel)
    assert_geometry(scene_view, proxy_map, sm.ImageGraphModel)
    assert_geometry(scene_view, proxy_map, sm.WebCamGraphModel)
    assert_geometry(scene_view, proxy_map, sm.VectorRollGraphModel)


def test_scatter_graph_widgets(setup_load_scene_model):
    proxy_map, scene_view = setup_load_scene_model
    assert_geometry(scene_view, proxy_map, sm.ScatterGraphModel)
    assert_geometry(scene_view, proxy_map, sm.VectorScatterGraphModel)


def test_text_widgets(setup_load_scene_model):
    proxy_map, scene_view = setup_load_scene_model
    assert_geometry(scene_view, proxy_map, sm.LabelModel)
    assert_geometry(scene_view, proxy_map, sm.StickerModel)


def test_instance_widget(setup_load_scene_model):
    proxy_map, scene_view = setup_load_scene_model
    assert_geometry(scene_view, proxy_map, sm.InstanceStatusModel,
                    keys=["DEVICE1.deviceId"])


@pytest.mark.skipif(IS_MAC_SYSTEM, reason="Link widgets are flaky on MacOS")
def test_link_widgets(setup_load_scene_model):
    proxy_map, scene_view = setup_load_scene_model
    assert_geometry(scene_view, proxy_map, sm.DeviceSceneLinkModel)
    assert_geometry(scene_view, proxy_map, sm.SceneLinkModel)
    assert_geometry(scene_view, proxy_map, sm.WebLinkModel)


def test_tool_widgets(setup_load_scene_model):
    proxy_map, scene_view = setup_load_scene_model
    assert_geometry(scene_view, proxy_map, sm.ImageRendererModel,
                    image=sm.create_base64image("svg", b"karabo"))


def test_trend_graph_widgets(setup_load_scene_model):
    proxy_map, scene_view = setup_load_scene_model
    assert_geometry(scene_view, proxy_map, sm.AlarmGraphModel)
    assert_geometry(scene_view, proxy_map, sm.StateGraphModel)
    assert_geometry(scene_view, proxy_map, sm.TrendGraphModel)


def test_vector_graph_widgets(setup_load_scene_model):
    proxy_map, scene_view = setup_load_scene_model
    assert_geometry(scene_view, proxy_map, sm.MultiCurveGraphModel)
    assert_geometry(scene_view, proxy_map, sm.NDArrayGraphModel)
    assert_geometry(scene_view, proxy_map, sm.VectorGraphModel)
    assert_geometry(scene_view, proxy_map, sm.VectorHistGraphModel)
    assert_geometry(scene_view, proxy_map, sm.VectorBarGraphModel)
    assert_geometry(scene_view, proxy_map, sm.VectorFillGraphModel)
    assert_geometry(scene_view, proxy_map, sm.VectorXYGraphModel)

# -----------------------------------------------------------------------------
# Clipboard actions test:


def test_widgets(setup_scene_view):
    """These are standalone widgets that are not bound to any devices."""
    proxy_map, scene_view, _ = setup_scene_view
    # Test single model selection
    label_model = get_label_model("foo")
    pos = (30, 30)
    with setup_and_test_selection(proxy_map, scene_view, label_model,
                                  pos=pos) as scene_model:
        SceneCopyAction().perform(scene_view)
        with setup_paste_action(scene_view, pos):
            ScenePasteAction().perform(scene_view)
    assert_copied_models(scene_view, scene_model, label_model)

    # Test single layout selection
    layout_model = get_layout_model(label_model)
    assert_clipboard_action(proxy_map, scene_view, layout_model)

    # Test multiple models selection
    another_label = get_label_model("bar", x=10, y=10)
    assert_clipboard_action(proxy_map, scene_view, label_model, another_label)

    # Test multiple layouts selection
    another_layout = get_layout_model(another_label, x=10, y=10)
    assert_clipboard_action(proxy_map, scene_view, layout_model,
                            another_layout)


def test_shapes(setup_scene_view):
    proxy_map, scene_view, _ = setup_scene_view
    # Test single line shape
    line_model = sm.LineModel(x1=10, y1=20, x2=100, y2=100)
    assert_clipboard_action(proxy_map, scene_view, line_model)

    # Test single arrow shape
    arrow_model = sm.ArrowPolygonModel(x1=30, y1=40, x2=300, y2=400)
    assert_clipboard_action(proxy_map, scene_view, arrow_model)

    # Test multiple shapes
    assert_clipboard_action(proxy_map, scene_view, line_model, arrow_model)
